unit jdxUnit1;
(*
This software is published under the terms of the GNU General Public
License (GPL, see below). For a detailed description of this license,
see http://www.gnu.org/copyleft/gpl.html

  JDXview - a utility program for displaying spectra in JCAMP-DX format.
  Author: Dr. Norbert Haider, University of Vienna, Department of Drug
  Synthesis, Althanstrasse 14, A-1090 Vienna, Austria.

  Inspired by Dr. Robert Lancashire's JCAMP-DX data viewer and its
  reincarnation in the MDL Chime plug-in, JDXview was developed as an
  external "helper application" which can be used in combination with
  web browsers, if they do not have a suitable plug-in installed.
  JDXview displays various kinds of spectra (NMR, IR, MS, etc.) in
  JCAMP-DX format. Supported formats are: uncompressed (AFFN) and
  compressed (ASDF, e.g. DIFDUP) as described in the 4.24 specification.
  Some features of more recent JCAMP-DX versions are also implemented,
  but not all of them.

  Furthermore, JDXview offers also rudimentary processing options: the
  range of a spectrum can be clipped, and the result saved in JCAMP-DX
  format again. Zooming and measuring of distances can be performed by
  two cursors which are placed by pressing the left or right mouse button,
  respectively, followed by a click on the "< >" button.

  Graphics output is supported in the following vector graphics formats:
  HP-GL (Hewlett-Packard Graphics Language)
  SVG (Scalable Vector Graphics)
  EPS (Encapsulated PostScript)

===============================================================================
DISCLAIMER
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
===============================================================================
*)


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, Menus, StdCtrls;

(* Kylix version:
uses
  SysUtils, Types, Classes, Variants, QGraphics, QControls, QForms, QDialogs,
  QMenus, QTypes, QExtCtrls, QComCtrls, QStdCtrls;
*)

type
  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    Exit1: TMenuItem;
    Spectrum1: TMenuItem;
    Parameters1: TMenuItem;
    N1: TMenuItem;
    OpenDialog1: TOpenDialog;
    StatusBar1: TStatusBar;
    Bevel1: TBevel;
    ScrollBox1: TScrollBox;
    PaintBox1: TPaintBox;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    Button4: TButton;
    Button5: TButton;
    LscrLbl: TLabel;
    RscrLbl: TLabel;
    Nextpage1: TMenuItem;
    SaveDialog1: TSaveDialog;
    Save1: TMenuItem;
    Clipspectrum1: TMenuItem;
    help1: TMenuItem;
    About1: TMenuItem;
    Plot1: TMenuItem;
    HPGLfile1: TMenuItem;
    SVGfile1: TMenuItem;
    EPSfile1: TMenuItem;
    Grid1: TMenuItem;
    procedure Exit1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure PaintBox1Paint(Sender: TObject);
    procedure Parameters1Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure DumpXYdata1Click(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Nextpage1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure Clipspectrum1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure HPGLfile1Click(Sender: TObject);
    procedure SVGfile1Click(Sender: TObject);
    procedure EPSfile1Click(Sender: TObject);
    procedure Grid1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure DrawCursor(CX : integer; cno : integer);
  end;

const
  debug  : boolean   = true;
  dtXYdata           =  101;
  dtXYpoints         =  102;
  dtpeaktable        =  103;
  dtntuples          =  104;
  stIR               = 1100;
  stNMR              = 1200;
  stUV               = 1300;
  stMS               = 1400;
  stRaman            = 1500;
  stUnknown          = 1900;
  sstIRtransmittance = 1101;
  sstIRabsorbance    = 1102;
  sstIRreflectance   = 1103;
  sstIRgeneric       = 1104;
  sstNMRfid          = 1201;
  sstNMRreal         = 1202;
  sstNMRcomplex      = 1203;
  sstNMRpeaktable    = 1204;
  sstNMRgeneric      = 1205;
  sstUVtransmittance = 1301;
  sstUVabsorbance    = 1302;
  sstUVgeneric       = 1303;
  sstMSpeaklist      = 1401;
  sstMSgeneric       = 1402;
  dmAFFN             = 2001;
  dmASDF             = 2002;

var
  Form1             : TForm1;
  jdxstringlist1    : Tstringlist;
  jdxstringlist2    : Tstringlist;
  jdxfilename       : string;
  datatype          : integer;
  spectype          : integer;
  specsubtype       : integer;
  specloaded        : boolean;
  datamode          : integer;
  loclipval         : double;
  hiclipval         : double;
  loclipstr         : string;
  hiclipstr         : string;

implementation

uses jdxUnit2, jdxUnit3;

{$R *.DFM}

const
  AppName              = 'JDXview';
  AppVersion           = 'v0.2d';
  nographics : boolean = false;
  maxpoints            = 65536;   // was: 32768
  ETX                  =  #3;
  CtrlD                =  #4;
  TAB                  =  #9;
  LF                   = #10;
  CR                   = #13;
  SO                   = #14;
  SI                   = #15;
  ln10      : double   = 2.302585093;
  blankstr  : string   = ' '+tab;
  junkstr   : string   = ' -_/\'+TAB;
  posSQZstr : string   = '@ABCDEFGHI';
  negSQZstr : string   = 'abcdefghi';
  posDIFstr : string   = '%JKLMNOPQR';
  negDIFstr : string   = 'jklmnopqr';
  DUPstr    : string   = 'STUVWXYZs';
  dc        : integer  = 1;  // direction coefficient
  maxtuples            = 8;

type
  xyrec = record
            X : single;
            Y : single;
          end;
  xyarray = array[1..maxpoints] of xyrec;
  jdxrec  = record
              title        : string;
              spectype     : string;
              origin       : string;
              owner        : string;
              date         : string;
              time         : string;
              spectrometer : string;
              params       : string;
              xunits       : string;
              yunits       : string;
              xfactor      : extended;
              yfactor      : extended;
              firstx       : extended;
              lastx        : extended;
              npoints      : integer;
              firsty       : extended;
              maxy         : extended;
              miny         : extended;
              xydata       : string;
              xypoints     : string;
              peaktable    : string;
              ntuples      : string;
              ntcount      : integer;
            end;
  ntuplerec = record
              var_name     : string;
              symbol       : string;
              var_type     : string;
              var_form     : integer;    // either dmAFFN or dmASDF
              var_dim      : longint;
              units        : string;
              first        : double;
              last         : double;
              min          : double;
              max          : double;
              factor       : double;
            end;
  ntstrtype = array[1..maxtuples] of string;

var
  xydata        : ^xyarray;
  jdxrec1       : jdxrec;
  maxt, mint    : extended;
  maxX, maxY    : longint;
  maxXval       : double;
  minXval       : double;
  offsetleft    : longint;
  offsetbottom  : longint;
  baseX, fullX  : longint;
  baseY,fullY   : longint;
  offsetXval    : extended;
  offsetYval    : extended;
  pdc           : integer;  // plot direction coefficient (+1 or -1)
  maxYfound     : extended;
  minYfound     : extended;
  maxYadj       : extended;
  minYadj       : extended;
  YscaleLo      : longint;
  YscaleHi      : longint;
  Yscalefactor  : double;
  Yscaledigits  : integer;
  XscaleLo      : longint;
  XscaleHi      : longint;
  iniXscaleLo   : longint;
  iniXscaleHi   : longint;
  Xscalefactor  : double;
  Xscaledigits  : integer;
  Xscalemod     : integer;
  XLo           : double;
  XHi           : double;
  iniXLo        : double;
  iniXHi        : double;
  iniXscalefactor: double;
  LcursX        : longint;
  LcursXval     : double;
  LcursXindex   : longint;
  LcursYval     : double;
  Lcdrawing     : boolean;
  RcursX        : longint;
  RcursXval     : double;
  RcursYval     : double;
  RcursXindex   : longint;
  Rcdrawing     : boolean;
  inscrollbox   : boolean;
  XindexLo      : longint;
  XindexHi      : longint;
  iniXindexLo   : longint;
  iniXindexHi   : longint;
  li            : longint;
  SFO1          : double;
  OFFSET        : double;
  REFPOINT      : double;
  dppm          : double;
  ppmXFACTOR    : double;
  ppmFIRSTX     : double;
  ppmLASTX      : double;
  ppmDELTAX     : double;
  foundHZ       : boolean;
  foundPPM      : boolean;
  foundSFO1     : boolean;
  foundOFFSET   : boolean;
  foundREFPOINT : boolean;
  showdeltaHz   : boolean;
  deltaHz       : double;
  NMRnucleus    : string;
  NMRscans      : integer;
  grid          : boolean;
  xdecimals     : integer;
//====new in version 0.1d
  foundShiftRef : boolean;
  shiftref_OK   : boolean;
  shiftref_mode : string;
  shiftref_solv : string;
  shiftref_point: longint;
  shiftref_ppm  : double;
  nt            : array[1..maxtuples] of ntuplerec;
  ntstr         : ntstrtype;
  foundntuples  : boolean;
  tuples        : integer;
  page          : integer;  // required for ntuples display

//========general functions and procedures=========

procedure debugoutput(dstr:string);
begin
  messagedlg(dstr,mtInformation,[mbOK],0);
end;

procedure displaywarning(wstr:string);
begin
  messagedlg(wstr,mtWarning,[mbOK],0);
end;

procedure upstr(var lstr:string);
var
  ii : integer;
begin
  if lstr <> '' then for ii := 1 to length(lstr) do
    lstr[ii] := upcase(lstr[ii]);
end;

//========JCAMP-DX header parsing routines=========

procedure ntsplitstr(instr:string; var ntlstr:ntstrtype);
const
  numstr : string = '+-.0123456789eE';
var
  auxstr : string;
  ni : integer;
begin
  instr := trim(instr);
  while (length(instr) > 0) and (instr[1]='(') do delete(instr,1,1);
  while (length(instr) > 0) and (instr[length(instr)]=')') do delete(instr,length(instr),1);
  ni := 0;
  while (length(instr) >0) and (ni < maxtuples) do
    begin
      inc(ni);
      auxstr := '';
      while (length(instr) > 0) and (instr[1]<>',') do
        begin
          if instr[1] <> ' ' then auxstr := auxstr + instr[1];
          delete(instr,1,1);
        end;
      ntlstr[ni] := auxstr;
      while (length(instr) > 0) and (instr[1]=',') do delete(instr,1,1);  // remove comma
    end;
end;

function jdxparnamestr(instr:string):string;
var
  ii : integer;
  auxstr  : string;
begin
  auxstr := '';
  if (instr <> '') and (pos('##',instr)=1) and (pos('=',instr)>3) then
    begin
      delete(instr,1,2);
      for ii := 1 to pos('=',instr)-1 do
        begin
          if pos(instr[ii],junkstr)=0 then auxstr := auxstr + instr[ii];
        end;
      upstr(auxstr);
    end;
  jdxparnamestr := auxstr;
end;

function jdxparvaluerawstr(instr:string):string;
var
  auxstr  : string;
begin
  auxstr := '';
  if (instr <> '') and (pos('##',instr)=1) and (pos('=',instr)>3) then
    delete(instr,1,pos('=',instr));
  auxstr := instr;
  jdxparvaluerawstr := auxstr;
end;

function jdxparvaluestr(instr:string):string;
var
  auxstr   : string;
begin
  auxstr := jdxparvaluerawstr(instr);
  if pos('$$',auxstr)>0 then auxstr := copy(auxstr,1,(pos('$$',auxstr)-1));
  while (length(auxstr)>0) and (pos(auxstr[1],blankstr)>0) do delete(auxstr,1,1);
  while (length(auxstr)>0) and (pos(auxstr[length(auxstr)],blankstr)>0) do delete(auxstr,length(auxstr),1);
  upstr(auxstr);
  jdxparvaluestr := auxstr;
end;

function jdxparvalueint(instr:string):longint;
var
  ii : integer;
  code : integer;
  auxstr  : string;
  nstr    : string;
  auxlong : longint;
  intnumstr : string;
begin
  auxlong := 0;
  intnumstr := '+-0123456789';
  auxstr := jdxparvaluestr(instr);
  if auxstr <> '' then
    begin
      nstr := '';
      for ii := 1 to length(auxstr) do
        if pos(auxstr[ii],intnumstr)>0 then nstr := nstr + auxstr[ii];
      // last character must be a digit!
      while (length(auxstr)>0) and (pos(auxstr[length(auxstr)],'0123456789')=0) do delete(auxstr,length(auxstr),1);
      val(nstr,auxlong,code);
      if (code <> 0) and debug then debugoutput('Could not get integer value from '+instr);
    end;
  jdxparvalueint := auxlong;
end;

function jdxparvaluefloat(instr:string):double;
var
  ii : integer;
  code : integer;
  auxstr  : string;
  nstr    : string;
  auxfloat : double;
  floatnumstr : string;
begin
  auxfloat := 0;
  floatnumstr := '+-.0123456789eE';
  auxstr := jdxparvaluestr(instr);
  if auxstr <> '' then
    begin
      nstr := '';
      for ii := 1 to length(auxstr) do
        if pos(auxstr[ii],floatnumstr)>0 then nstr := nstr + auxstr[ii];
      // last character must be a digit!
      while (length(auxstr)>0) and (pos(auxstr[length(auxstr)],'0123456789')=0) do delete(auxstr,length(auxstr),1);
      val(nstr,auxfloat,code);
      if (code <> 0) and debug then debugoutput('Could not get float value from '+instr+' nstr: '+nstr);
    end;
  jdxparvaluefloat := auxfloat;
end;

procedure initjdxrec;
begin
  with jdxrec1 do
    begin
      title        := '';
      spectype     := '';
      origin       := '';
      owner        := '';
      date         := '';
      time         := '';
      spectrometer := '';
      params       := '';
      xunits       := '';
      yunits       := '';
      xfactor      := 1;
      yfactor      := 1;
      firstx       := 1;
      lastx        := 1;
      npoints      := 0;
      firsty       := 0;
      maxy         := 0;
      miny         := 0;
      xydata       := '';
      xypoints     := '';
      peaktable    := '';
      ntuples      := '';
      ntcount      := 0;
    end;
end;

procedure initNMRparams;
begin
  SFO1          := 0;
  OFFSET        := 0;
  REFPOINT      := 0;
  dppm          := 0;
  ppmXFACTOR    := 1;
  ppmFIRSTX     := 0;
  ppmLASTX      := 0;
  ppmDELTAX     := 0;
  foundHZ       := false;
  foundPPM      := false;
  foundSFO1     := false;
  foundOFFSET   := false;
  foundREFPOINT := false;
  showdeltaHz   := false;
  deltaHz       := 0;
  NMRnucleus    := '';
  NMRscans      := 0;
  //==== new in v0.1d
  foundShiftRef := false;
  shiftref_OK   := false;
  shiftref_mode := '';
  shiftref_solv := '';
  shiftref_point:= 0;
  shiftref_ppm  := 0;
end;

procedure adjustminmaxX;
var
  coeff   : double;
  intsign : integer;
  margin  : double;
  minXadj, maxXadj : integer;
begin
  // calculate the values for the X scale
  coeff := 1;
  while coeff*abs(XLo - XHi) < 5  do coeff := coeff * 10;
  while coeff*abs(XLo - XHi) > 50 do coeff := coeff / 10;
  if (XLo >= 0) then
    begin
      intsign := 1;
      if XHi < XLo then margin  := 0.5 else margin := -0.5;
    end else
    begin
      intsign := -1;
      if XHi < XLo then margin  := -0.5 else margin := 0.5;
    end;
  minXadj := round(intsign*abs(XLo)*coeff - margin);
  if (XHi >= 0) then
    begin
      intsign := 1;
      if XHi < XLo then margin  := 0.5 else margin := -0.5;
    end else
    begin
      intsign := -1;
      if XHi < XLo then margin  := 0.5 else margin := -0.5;
    end;
  if (spectype = stMS) and (jdxrec1.peaktable <> '') then margin := 0;
  maxXadj := round(intsign*abs(XHi)*coeff - margin);
  XscaleHi := round(maxXadj);
  XscaleLo := round(minXadj);
  Xscalefactor := coeff;
  if abs(maxXadj-minXadj)/Xscalefactor > 50 then  Xscaledigits := 0 else
    if abs(maxXadj-minXadj)/Xscalefactor > 5 then  Xscaledigits := 1 else
      Xscaledigits := 2;
  if (spectype = stIR) or (spectype = stNMR) then if XLo > XHi then pdc := 1 else pdc := -1;
  if (specsubtype = sstNMRfid) then if XLo > XHi then pdc := -1 else pdc := 1;
  if (spectype = stMS) or (spectype = stUV) then if XLo < XHi then pdc := 1 else pdc := -1;
end;

procedure adjustminmaxY;
var
  coeff   : double;
  intsign : integer;
  margin  : double;
begin
  maxYadj := maxYfound;
  minYadj := minYfound;
  coeff := 1;
  while coeff*abs(maxYadj-minYadj) < 2 do coeff := coeff * 10;
  while coeff*abs(maxYadj-minYadj) > 20 do coeff := coeff / 10;
  if maxYadj >= 0 then
    begin
      intsign := 1;
      margin  := 0.5;
    end else
    begin
      intsign := -1;
      margin  := -0.5;
    end;
  maxYadj := intsign*round(coeff*abs(maxYadj) + margin)/coeff;
  if minYadj >= 0 then
    begin
      intsign := 1;
      margin  := 0.5;
    end else
    begin
      intsign := -1;
      margin  := -0.5;
    end;
  minYadj := intsign*round(coeff*abs(minYadj) - margin)/coeff;
  if (spectype = stIR) and (specsubtype = sstIRtransmittance) then
    begin
      if maxYfound > 2 then
        begin
          maxYadj := 100;
          if minYfound <= 60 then minYadj := 0;
        end;
      if maxYfound < 1 then
        begin
          maxYadj := 1;
          if minYfound <= 0.6 then minYadj := 0;
        end;
    end;
  if (spectype = stMS) and (jdxrec1.peaktable <> '') then
    begin
      maxYadj := 105;
      minYadj := 0;
    end;
  if maxYadj = minYadj then maxYadj := minYadj + 1;   // just to be sure
  // now calculate the values for the Y scale
  YscaleHi := round(maxYadj*coeff);
  YscaleLo := round(minYadj*coeff);
  Yscalefactor := coeff;
  Yscaledigits := 0;
  if abs(YscaleHi-YscaleLo)/coeff <  10 then Yscaledigits := 1;
  if abs(YscaleHi-YscaleLo)/coeff <   1 then Yscaledigits := 2;
  if abs(YscaleHi-YscaleLo)/coeff < 0.1 then Yscaledigits := 3;  // preliminary...
end;

procedure adjustminmax;
begin
  adjustminmaxX;
  adjustminmaxY;
  iniXLo := XLo;
  iniXHi := XHi;
  XindexLo := 1;
  XindexHi := jdxrec1.npoints;
  iniXscaleLo := XscaleLo;
  iniXscaleHi := XscaleHi;
  iniXindexLo := XindexLo;
  iniXindexHi := XindexHi;
  iniXscalefactor := Xscalefactor;
end;

procedure process_shiftref(srstr:string);  // new in v0.1d
const
  numstr : string = '+-.0123456789eE';
var
  auxstr : string;
  auxint : longint;
  auxdbl : double;
  code   : integer;
begin
  auxstr := '';
  auxint := 0;
  auxdbl := 0;
  shiftref_OK := true;
  // remove leading and trailing whitespace
  srstr := trim(srstr);
  // remove enclosing brackets (which should not be there according to specs, but...)
  while (length(srstr) > 0) and (srstr[1]='(') do delete(srstr,1,1);
  while (length(srstr) > 0) and (srstr[length(srstr)]=')') do delete(srstr,length(srstr),1);
  // read 1st parameter (string, either 'INTERNAL' or 'EXTERNAL')
  while (length(srstr) > 0) and (srstr[1]<>',') do
    begin
      if srstr[1] <> ' ' then auxstr := auxstr + srstr[1];
      delete(srstr,1,1);
    end;
  shiftref_mode := auxstr;
  auxstr := '';
  while (length(srstr) > 0) and (srstr[1]=',') do delete(srstr,1,1);  // remove comma
  // read 2nd parameter (text, name of the reference compound, e.g. 'CDCl3')
  while (length(srstr) > 0) and (srstr[1]<>',') do
    begin
      if srstr[1] <> ' ' then auxstr := auxstr + srstr[1];
      delete(srstr,1,1);
    end;
  shiftref_solv := auxstr;
  auxstr := '';
  while (length(srstr) > 0) and (srstr[1]=',') do delete(srstr,1,1);  // remove comma
  // read 3rd parameter (AFFN (integer), data point number, between 1 and NPOINTS)
  while (length(srstr) > 0) and (srstr[1]<>',') do
    begin
      if (pos(srstr[1],numstr)>0) then auxstr := auxstr + srstr[1];
      delete(srstr,1,1);
    end;
  val(auxstr,auxint,code);
  if (code = 0) then shiftref_point := auxint else shiftref_OK := false;
  auxstr := '';
  while (length(srstr) > 0) and (srstr[1]=',') do delete(srstr,1,1);  // remove comma
  // read 4th parameter (AFFN (float), ppm value of reference point)
  while (length(srstr) > 0) do
    begin
      if (pos(srstr[1],numstr)>0) then auxstr := auxstr + srstr[1];
      delete(srstr,1,1);
    end;
  val(auxstr,auxdbl,code);
  if (code = 0) then shiftref_ppm := auxdbl else shiftref_OK := false;
end;

procedure calcnewNMRparams;
var
  sw_ppm : double;
begin
  if SFO1 = 0 then
    begin
      debugoutput('Spectrometer frequency not found');
      dppm := 0;
      exit;
    end;
  sw_ppm := (jdxrec1.FIRSTX - jdxrec1.LASTX)/SFO1;
  if foundShiftRef and shiftref_OK then
    begin
      debugoutput('Using shift reference!');
      if shiftref_point < 1 then shiftref_point := 1; // correction for Bruker bug
      OFFSET := shiftref_ppm + sw_ppm * (shiftref_point-1)/(jdxrec1.NPOINTS-1);
    end else
    begin
      if not foundOFFSET then
        begin
          if not foundREFPOINT then
            begin
              //debugoutput('Reference point not found');
              dppm := 0;
              exit;
            end;
          OFFSET := (abs(jdxrec1.LASTX-jdxrec1.FIRSTX)-REFPOINT)/SFO1;
        end;
    end;
  dppm      := (jdxrec1.FIRSTX/SFO1)-OFFSET;
  ppmFIRSTX := OFFSET;
//  ppmLASTX  := jdxrec1.LASTX/SFO1 - dppm;
  ppmLASTX  := OFFSET - sw_ppm;
//  ppmDELTAX := (ppmLASTX-ppmFIRSTX)/(jdxrec1.NPOINTS-1);
  ppmDELTAX := sw_ppm;
  ppmXFACTOR:= 1;
end;

function hz2ppm:boolean;
var
  retval : boolean;
  i : integer;
  d1, d2 : extended;
begin
  retval := false;
  maxXval := -10000;
  minXval :=  10000;
  for i := 1 to jdxrec1.npoints do
    begin
      d1 := xydata^[i].X;
      d2 := d1/SFO1 - dppm;
      xydata^[i].X := d2;
      if maxXval < d2 then maxXval := d2;
      if minXval > d2 then minXval := d2;
    end;
  ppmfirstx := xydata^[1].X;                 // v0.2d
  ppmlastx  := xydata^[jdxrec1.npoints].X;   // v0.2d
  hz2ppm := retval;
end;

procedure updateNMRparams;
begin
  with jdxrec1 do
    begin
      xunits  := 'PPM';
      firstX  := ppmfirstX;
      lastX   := ppmlastX;
      xfactor := ppmxfactor;
    end;
end;

procedure processntuples(p:integer);
var
  i, n : integer;
begin
  n := 0;
  for i := 1 to maxtuples do if nt[i].var_name <> '' then inc(n);
  tuples := n;
  if (nt[1].symbol = 'X') and (nt[1].var_dim <= maxpoints) then
    begin
      jdxrec1.npoints  := nt[1].var_dim;
      jdxrec1.firstx  := nt[1].first;
      jdxrec1.lastx   := nt[1].last;
      jdxrec1.xfactor := nt[1].factor;
      jdxrec1.xunits  := nt[1].units;
      if (nt[(p+1)].var_type = 'DEPENDENT') then
        begin
          jdxrec1.firsty  := nt[(p+1)].first;
          jdxrec1.yfactor := nt[(p+1)].factor;
          jdxrec1.miny    := nt[(p+1)].min;
          jdxrec1.maxy    := nt[(p+1)].max;
          datamode        := nt[(p+1)].var_form;
        end;
    end;
    calcnewNMRparams;
end;

function readjdx:boolean;
var
  i,s       : integer;
  tmpstr    : string;
  outstr    : string;
  combstr   : string;
  xcalcd    : double;
  xfound    : double;
  nvalues   : integer;
  lastvalue : double;
  dif       : boolean;
  prevdif   : boolean;
  auxreal   : double;

  function parsejdxntuplesline(ntline:string):string;
  var
    auxstr : string;
    auxint : longint;
    auxdbl : double;
    code : integer;
    i : integer;
  begin
    for i := 1 to maxtuples do ntstr[i] := '';
    if jdxparnamestr(ntline)='VAR_NAME' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do nt[i].var_name := ntstr[i];
      end;
    if jdxparnamestr(ntline)='SYMBOL' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do nt[i].symbol := ntstr[i];
      end;
    if jdxparnamestr(ntline)='VARTYPE' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do nt[i].var_type := ntstr[i];
      end;
    if jdxparnamestr(ntline)='VARFORM' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].var_form := 0;
            if ntstr[i] = 'AFFN' then nt[i].var_form := dmAFFN;
            if ntstr[i] = 'ASDF' then nt[i].var_form := dmASDF;
          end;
      end;
    if jdxparnamestr(ntline)='VARDIM' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].var_dim := 0;
            val(ntstr[i],auxint,code);
            if (code = 0) then nt[i].var_dim := auxint;
          end;
      end;
    if jdxparnamestr(ntline)='UNITS' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do nt[i].units := ntstr[i];
      end;
    if jdxparnamestr(ntline)='FIRST' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].first := 0;
            val(ntstr[i],auxdbl,code);
            if (code = 0) then nt[i].first := auxdbl;
          end;
      end;
    if jdxparnamestr(ntline)='LAST' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].last := 0;
            val(ntstr[i],auxdbl,code);
            if (code = 0) then nt[i].last := auxdbl;
          end;
      end;
    if jdxparnamestr(ntline)='MIN' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].min := 0;
            val(ntstr[i],auxdbl,code);
            if (code = 0) then nt[i].min := auxdbl;
          end;
      end;
    if jdxparnamestr(ntline)='MAX' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].max := 0;
            val(ntstr[i],auxdbl,code);
            if (code = 0) then nt[i].max := auxdbl;
          end;
      end;
    if jdxparnamestr(ntline)='FACTOR' then
      begin
        auxstr := jdxparvaluestr(ntline);
        ntsplitstr(auxstr,ntstr);
        for i := 1 to maxtuples do
          begin
            nt[i].factor := 0;
            val(ntstr[i],auxdbl,code);
            if (code = 0) then nt[i].factor := auxdbl;
          end;
      end;
    parsejdxntuplesline := '';
  end;

  function ntuples_count:integer;
  var
    i, n : integer;
  begin
    n := 0;
    for i := 1 to maxtuples do if nt[i].var_name <> '' then inc(n);
    ntuples_count := n;
  end;

  function parsejdxheaderline(hline:string):string;
  var
    ii, jj : integer;
    errstr : string;
    auxstr : string;
    // remove comment in Kylix version:
    //junkstr : string;
    parline : string;
  begin
    errstr := '';
    auxstr := '';
    parline := '';
    junkstr := ' -_\/';
    ii := pos('=',hline);
    if ii = 0 then
      begin
        parsejdxheaderline := '';
        exit;
      end;
    for jj := 1 to ii do
      if pos(hline[jj],junkstr)=0 then parline := parline + upcase(hline[jj]);
    if jdxparnamestr(hline)='TITLE' then
      begin
        jdxrec1.title := jdxparvaluerawstr(hline);
        jdxrec1.title := trimleft(jdxrec1.title);
      end;
    if jdxparnamestr(hline)='DATATYPE' then
      begin
        auxstr := jdxparvaluestr(hline);
        if auxstr <> '' then jdxrec1.spectype := auxstr;
        spectype := stUnknown;
        if (pos('INFRARED',auxstr)>0) then spectype := stIR;
        if (pos('NMR ',auxstr)>0) or (pos('NUCLEAR MAGNETIC',auxstr)>0) then spectype := stNMR;
        if (pos('UV ',auxstr)>0) or (pos('UV/V',auxstr)>0) or (pos('ULTRAVIOLET',auxstr)>0) then spectype := stUV;
        if (pos('MS ',auxstr)>0) or (pos('MASS',auxstr)>0) then spectype := stMS;
        if pos('RAMAN',auxstr)>0 then spectype := stRaman;
      end;
    if jdxparnamestr(hline)='ORIGIN' then jdxrec1.origin := jdxparvaluerawstr(hline);
    if jdxparnamestr(hline)='OWNER'  then jdxrec1.owner  := jdxparvaluerawstr(hline);
    if jdxparnamestr(hline)='DATE'   then jdxrec1.date   := jdxparvaluestr(hline);
    if jdxparnamestr(hline)='TIME'   then jdxrec1.time   := jdxparvaluestr(hline);
    if jdxparnamestr(hline)='SPECTROMETER' then jdxrec1.spectrometer   := jdxparvaluerawstr(hline);
    if jdxparnamestr(hline)='INSTRUMENTPARAMETERS' then jdxrec1.params := jdxparvaluerawstr(hline);
    if jdxparnamestr(hline)='XUNITS' then jdxrec1.xunits := jdxparvaluestr(hline);
    if jdxparnamestr(hline)='YUNITS' then
      begin
        jdxrec1.yunits := jdxparvaluestr(hline);
        upstr(jdxrec1.yunits);
        if pos('TRANSM',jdxrec1.yunits)>0 then
          begin
            if spectype = stIR then specsubtype := sstIRtransmittance;
            if spectype = stUV then specsubtype := sstUVtransmittance;
          end;
        if pos('ABSORB',jdxrec1.yunits)>0 then
          begin
            if spectype = stIR then specsubtype := sstIRabsorbance;
            if spectype = stUV then specsubtype := sstUVabsorbance;
          end;
      end;
    if jdxparnamestr(hline)='XFACTOR' then
      begin
        jdxrec1.xfactor := jdxparvaluefloat(hline);
        if jdxrec1.xfactor = 0 then
          begin
            errstr := 'Invalid XFACTOR (0): '+hline;
            debugoutput(errstr);
            jdxrec1.xfactor := 1;
          end;
      end;
    if jdxparnamestr(hline)='YFACTOR' then
      begin
        jdxrec1.yfactor := jdxparvaluefloat(hline);
        if jdxrec1.yfactor = 0 then
          begin
            errstr := 'Invalid YFACTOR (0): '+hline;
            debugoutput(errstr);
            jdxrec1.yfactor := 0.00001;
          end;
      end;
    if jdxparnamestr(hline)='FIRSTX'   then jdxrec1.firstx   := jdxparvaluefloat(hline);
    if jdxparnamestr(hline)='LASTX'    then jdxrec1.lastx    := jdxparvaluefloat(hline);
    if jdxparnamestr(hline)='NPOINTS'  then jdxrec1.npoints  := jdxparvalueint(hline);
    if jdxparnamestr(hline)='FIRSTY'   then jdxrec1.firsty   := jdxparvaluefloat(hline);
    if jdxparnamestr(hline)='MAXY'     then jdxrec1.maxy     := jdxparvaluefloat(hline);
    if jdxparnamestr(hline)='MINY'     then jdxrec1.miny     := jdxparvaluefloat(hline);
    if jdxparnamestr(hline)='XYDATA'   then jdxrec1.xydata   := jdxparvaluestr(hline);
    if jdxparnamestr(hline)='XYPOINTS' then
      begin
        jdxrec1.xypoints := jdxparvaluestr(hline);
        datatype := dtXYpoints;
      end;
    if jdxparnamestr(hline)='PEAKTABLE' then
      begin
        jdxrec1.peaktable := jdxparvaluestr(hline);
        datatype := dtpeaktable;
      end;
    if jdxparnamestr(hline)='NTUPLES' then
      begin
        jdxrec1.ntuples := jdxparvaluestr(hline);
        datatype := dtpeaktable;
        auxstr := jdxparvaluestr(hline);
        if (pos('NMR',auxstr)) > 0 then
          begin
            spectype := stNMR;
            if (pos('FID',auxstr)) > 0 then
              begin
                specsubtype := sstNMRfid;
                pdc := -1;
              end;
            if (pos('SPECTRUM',auxstr)) > 0 then
              begin
                specsubtype := sstNMRreal;
                pdc := 1;
              end;
          end;
      end;
    // watch out for NMR parameters
    auxstr := jdxparnamestr(hline);
    if pos('OBSERVEFREQUENCY',auxstr)>0 then SFO1 := jdxparvaluefloat(hline);
    if pos('$OFFSET',auxstr)>0 then
      begin
        foundOFFSET := true;
        OFFSET := jdxparvaluefloat(hline);
      end;
    if pos('HZ',jdxrec1.xunits)>0 then foundHZ := true;
    if pos('PPM',jdxrec1.xunits)>0 then foundPPM := true;
    if pos('$REFERENCEPOINT',auxstr)>0 then
      begin
        foundREFPOINT := true;
        REFPOINT := jdxparvaluefloat(hline);
      end;
    if pos('OBSERVENUCLEUS',auxstr)>0 then NMRnucleus := jdxparvaluestr(hline);
    if pos('AVERAGES',auxstr)>0 then NMRscans := jdxparvalueint(hline);
    //==== new in v0.1d: look for SHIFT REFERENCE label
    if pos('.SHIFTREFERENCE',auxstr)>0 then
      begin
        foundShiftRef := true;
        auxstr := jdxparvaluerawstr(hline);
        process_shiftref(auxstr);
      end;
    //==== read NTUPLES
    if jdxparnamestr(hline)='NTUPLES' then
      begin
        foundntuples := true;
        auxstr := jdxparvaluestr(hline);
        if auxstr <> '' then jdxrec1.spectype := auxstr;
        spectype := stUnknown;
        if (pos('INFRARED',auxstr)>0) then spectype := stIR;
        if (pos('NMR',auxstr)>0) or (pos('NUCLEAR MAGNETIC',auxstr)>0) then spectype := stNMR;
        if (pos('UV ',auxstr)>0) or (pos('UV/V',auxstr)>0) or (pos('ULTRAVIOLET',auxstr)>0) then spectype := stUV;
        if (pos('MS ',auxstr)>0) or (pos('MASS',auxstr)>0) then spectype := stMS;
        if pos('RAMAN',auxstr)>0 then spectype := stRaman;
        datatype := dtXYdata;  // preliminary!!!
      end;
    if foundntuples then errstr := parsejdxntuplesline(hline);
    parsejdxheaderline := errstr;
  end;  // parsejdxheaderline function

  function parsejdxXYdataline(dline:string):string;
  const
    numstr : string = '.0123456789';
    sepstr : string = ' ,;+-'+tab;
  var
    ii, jj, code : integer;
    auxstr       : string;
    nstr         : string;
    n2str        : string;
    r0, r1       : extended;
    d1           : double;
    isaffn       : boolean;
    isvalid      : boolean;
    vc           : integer; (* plus or minus *)
    commacount   : integer;
    valuecount   : integer;
    dupcount     : integer;
    difvalue     : double;
    firstchar    : char;
    itmp         : integer;
    tempstr       : string;
  begin
    combstr := posSQZstr+negSQZstr+posDIFstr+negDIFstr+DUPstr+'?';
    xcalcd := 0;
    xfound := 0;
    auxstr := '';
    vc     := 1;
    r0     := 0;
    r1     := 0;
    difvalue := 0;
    if dline = '' then
      begin
        parsejdxXYdataline := '';
        exit;
      end;
    // remove comments
    if pos('$$',dline)>0 then dline := copy(dline,1,(pos('$$',dline)-1));
    if dline = '' then
      begin
        parsejdxXYdataline := '';
        exit;
      end;
    isaffn := true;
    isvalid := true;
    for ii := 1 to length(dline) do
      begin
        if pos(dline[ii],negSQZstr+posSQZstr+negDIFstr+posDIFstr+DUPstr) > 0
          then isaffn := false;
      end;
    if not isaffn then
      begin
        nstr := '';
        valuecount := 0;
        while (pos(' ',dline)>0) or (pos(TAB,dline)>0) do
          for ii := 1 to length(dline) do
            if pos(dline[ii],' '+TAB)>0 then delete(dline,ii,1);
        ii := 0;
        repeat (* read 1st value ( = X)  *)
          inc(ii);
          nstr := nstr + dline[ii];
        until (ii=length(dline)) or (pos(dline[(ii+1)],numstr)=0);
        val(nstr,xfound,code); (* should contain X *);
        xfound := xfound*jdxrec1.xfactor;
        if code=0 then inc(valuecount) else
          begin
            auxstr := 'invalid X value: '+nstr;
            parsejdxXYdataline := auxstr;
            exit;
          end;
        nstr := '';
        dupcount := 1;
        while (ii < length(dline)) do  (* and now the rest... *)
          begin
            inc(ii);
            nstr := nstr + dline[ii];
            if (ii=length(dline)) or (pos(dline[(ii+1)],combstr)>0) then
              begin  (* processing *)
                inc(valuecount);
                if pos(nstr[1],combstr)=0 then
                  begin
                    auxstr := 'not an ASDF number: '+nstr+' '+inttostr(nvalues);;
                    parsejdxXYdataline := auxstr;
                    exit;
                  end else
                  begin
                    if pos('?',nstr)>0 then r1 := lastvalue else
                      begin
                        firstchar := nstr[1];
                        if pos(firstchar,posSQZstr)>0 then
                          begin
                            dupcount := 1;
                            dif   := false;
                            n2str := nstr;
                            itmp  := ord(firstchar)-64;
                            tempstr := '';
                            str(itmp:1,tempstr);
                            n2str[1] := tempstr[1];
                            val(n2str,r1,code);
                            if code <> 0 then
                              begin
                                auxstr := 'invalid SQZ number: '+n2str;
                                parsejdxXYdataline := auxstr;
                                exit;
                              end;
                          end;
                        if pos(firstchar,negSQZstr)>0 then
                          begin
                            dupcount := 1;
                            dif   := false;
                            n2str := nstr;
                            itmp  := ord(firstchar)-96;
                            tempstr := '';
                            str(itmp:1,tempstr);
                            n2str[1] := tempstr[1];
                            val(n2str,r1,code);
                            if code <> 0 then
                              begin
                                auxstr := 'invalid SQZ number: '+n2str;
                                parsejdxXYdataline := auxstr;
                                exit;
                              end;
                            r1 := abs(r1)*(-1);
                          end;
                        if pos(firstchar,posDIFstr)>0 then
                          begin
                            dif := true;
                            n2str := nstr;
                            if firstchar='%' then itmp := 0 else
                              itmp  := ord(firstchar)-73;
                            tempstr := '';
                            str(itmp:1,tempstr);
                            n2str[1] := tempstr[1];
                            val(n2str,difvalue,code);
                            if code <> 0 then
                              begin
                                auxstr := 'invalid DIF number: '+n2str;
                                parsejdxXYdataline := auxstr;
                                exit;
                              end;
                          end;
                        if pos(firstchar,negDIFstr)>0 then
                          begin
                            dif := true;
                            n2str := nstr;
                            itmp  := ord(firstchar)-105;
                            tempstr := '';
                            str(itmp:1,tempstr);
                            n2str[1] := tempstr[1];
                            val(n2str,difvalue,code);
                            if code <> 0 then
                              begin
                                auxstr := 'invalid DIF number: '+n2str;
                                parsejdxXYdataline := auxstr;
                                exit;
                              end;
                            difvalue := abs(difvalue)*(-1);
                          end;
                        if pos(firstchar,DUPstr)>0 then
                          begin
                            n2str := nstr;
                            if firstchar='s' then itmp := 9 else
                              itmp  := ord(firstchar)-82;
                            tempstr := '';
                            str(itmp:1,tempstr);
                            n2str[1] := tempstr[1];
                            val(n2str,dupcount,code);
                            if dupcount > 1 then dec(dupcount);
                            if code <> 0 then
                              begin
                                auxstr := 'invalid DUP number: '+n2str;
                                parsejdxXYdataline := auxstr;
                                exit;
                              end;
                          end;
                      end;
                    if prevdif and (valuecount=2) then
                      begin (* Y value is only for checking... *)
                        if r1 <> lastvalue then
                          begin
                            auxstr := 'Y check failed: '+floattostr(lastvalue)+'/'
                              +floattostr(r1)+' at '+floattostr(xfound);
                            displaywarning(auxstr);
                            auxstr := '';
                            //parsejdxXYdataline := auxstr;
                            //exit;
                          end;
                        dif := false;
                      end else
                      begin
                        for jj := 1 to dupcount do
                          begin
                            if dif then r1 := lastvalue+difvalue;
                            xydata^[s].X := (jdxrec1.firstx -
                              (s-1)*((jdxrec1.firstx-jdxrec1.lastx)/(jdxrec1.npoints-1)));
                            xydata^[s].Y := r1*jdxrec1.yfactor*vc;
                            if xydata^[s].Y > maxt then maxt := xydata^[s].Y;
                            if xydata^[s].Y < mint then mint := xydata^[s].Y;
                            if xydata^[s].X > maxXval then maxXval := xydata^[s].X;
                            if xydata^[s].X < minXval then minXval := xydata^[s].X;
                            inc(nvalues);
                            inc(valuecount);
                            lastvalue := r1;
                            prevdif := dif;
                            if (s+dc >= 0) and (s+dc <= maxpoints) then s := s+dc;
                          end;
                        dupcount := 1;
                      end;
                  end;
                nstr := '';
              end;
          end;
      end else  // end ASDF
      begin     // begin AFFN
        for ii := 1 to length(dline) do
          begin
            if (pos(dline[ii],sepstr) = 0) and (pos(dline[ii],numstr) = 0)
              then isvalid := false;
          end;
        if not isvalid then
          begin
            auxstr := 'invalid data line '+dline;
            parsejdxXYdataline := auxstr;
            exit;
          end;
        nstr := '';
        valuecount := 0;
//        commacount := 0;
        ii := 1;
        while (ii < length(dline)) do
          begin
            commacount := 0;
            vc := 1;
            while (ii < length(dline)) and (pos(dline[ii],sepstr) > 0) do
              begin
                if dline[ii]=',' then inc(commacount);
                if dline[ii]='-' then vc := -1;
                if dline[ii]='+' then vc := 1;
                inc(ii);
              end;
            // last character in line must be a digit!
            while (length(dline)>0) and (pos(dline[length(dline)],'0123456789') = 0) do
              delete(dline,length(dline),1);
            if commacount > 1 then
              begin
                for jj := 2 to commacount do
                  begin
                    r1 := lastvalue;
                    d1 := r1*jdxrec1.yfactor;
                    xydata^[s].X := (jdxrec1.firstx -
                      (s-1)*((jdxrec1.firstx-jdxrec1.lastx)/(jdxrec1.npoints-1)));
                    xydata^[s].Y := d1*vc;
                    inc(nvalues);
                    inc(valuecount);
                    if (s+dc > 0) and (s+dc < maxpoints) then s := s+dc;
                  end;
              end;
            while (ii <= length(dline)) and (pos(dline[ii],numstr) > 0) do
              begin
                if ii <= length(dline) then nstr := nstr+dline[ii];
                inc(ii);
              end;
            val(nstr,r1,code);
            if (code=0) then
              begin
                inc(valuecount);
                nstr := '';
                if valuecount = 1 then  (* first value is X *)
                  begin
                    r0 := r1;
                  end;
                nstr := '';
                if valuecount > 1 then  (* first value is X *)
                  begin
                    if jdxrec1.xydata <> '' then
                      begin
                        xydata^[s].X := (jdxrec1.firstx -
                          (s-1)*((jdxrec1.firstx-jdxrec1.lastx)/(jdxrec1.npoints-1)));
                        xydata^[s].Y := r1*jdxrec1.yfactor*vc;
                      end;
                    if jdxrec1.peaktable <> '' then
                      begin
                        xydata^[s].X := jdxrec1.xfactor*r0;
                        xydata^[s].Y := r1*jdxrec1.yfactor*vc;
                      end;
                    if xydata^[s].Y > maxt then maxt := xydata^[s].Y;
                    if xydata^[s].Y < mint then mint := xydata^[s].Y;
                    if xydata^[s].X > maxXval then maxXval := xydata^[s].X;
                    if xydata^[s].X < minXval then minXval := xydata^[s].X;
                    inc(nvalues);
                    lastvalue := r1;
                    if (s+dc >= 0) and (s+dc <= maxpoints) then s := s+dc;
                  end;
              end else
              begin
                if nstr = '' then nstr := dline;
                auxstr := 'invalid data point: '+nstr;
                parsejdxXYdataline := auxstr;
                exit;
              end;
          end;
      end;
    parsejdxXYdataline := auxstr;
  end;

  function parsejdxXYpointline(dline:string):string;
  const
    numstr : string = '+-.0123456789eE';
    sepstr : string = ' ,;'+tab;
  var
    ii, code : integer;
    auxstr       : string;
    nstr         : string;
    isvalid      : boolean;
    valuecount   : integer;
    commacount   : integer;
    r1           : double;
  begin
    auxstr  := '';
    isvalid := true;
    if dline = '' then
      begin
        parsejdxXYpointline := '';
        exit;
      end;
    // remove comments
    if pos('$$',dline)>0 then dline := copy(dline,1,(pos('$$',dline)-1));
    if dline = '' then
      begin
        parsejdxXYpointline := '';
        exit;
      end;
    for ii := 1 to length(dline) do
      begin
        if (pos(dline[ii],sepstr) = 0) and (pos(dline[ii],numstr) = 0)
          then isvalid := false;
      end;
    if not isvalid then
      begin
        auxstr := 'invalid data line '+dline;
        parsejdxXYpointline := auxstr;
        exit;
      end;
        nstr := '';
        valuecount := 0;
        commacount := 0;
        ii := 1;
        while (ii < length(dline)) do
          begin
            commacount := 0;
            while (ii < length(dline)) and (pos(dline[ii],sepstr) > 0) do
              begin
                if dline[ii]=',' then inc(commacount);
                inc(ii);
              end;
            while (ii <= length(dline)) and (pos(dline[ii],numstr) > 0) do
              begin
                if ii <= length(dline) then nstr := nstr+dline[ii];
                inc(ii);
              end;
            val(nstr,r1,code);
            if (code=0) and (abs(r1)<37676*37676) then
              begin
                inc(valuecount);
                nstr := '';
                if odd(valuecount) then  (* odd value is X *)
                  begin
                    xydata^[s].X := r1*jdxrec1.xfactor;
                  end else
                  begin
                    xydata^[s].Y := r1*jdxrec1.yfactor;
                    if xydata^[s].Y > maxt then maxt := xydata^[s].Y;
                    if xydata^[s].Y < mint then mint := xydata^[s].Y;
                    if xydata^[s].X > maxXval then maxXval := xydata^[s].X;
                    if xydata^[s].X < minXval then minXval := xydata^[s].X;
                    inc(s);
                 end;
              end;
          end;
    parsejdxXYpointline := auxstr;
  end;

begin   // readjdx
  datatype := dtXYdata;
  jdxrec1.npoints := 0;
  if not (foundntuples and (page > 1)) then initNMRparams;
  specloaded := false;
  tmpstr := '';
  if not (li < (jdxstringlist1.count-2)) then
    begin
      readjdx := false;
      exit;
    end;
  repeat
    begin
      inc(li);
      tmpstr := jdxstringlist1.strings[li];
      outstr := parsejdxheaderline(tmpstr);
      if outstr <> '' then
        messagedlg('Parameter read error: '+outstr+' (line '+inttostr(li+1)+')',mtError,
        [mbOK],0);
    end;
  until (li = jdxstringlist1.count-1) or
        (jdxparnamestr(jdxstringlist1.strings[li]) = 'XYDATA') or
        (jdxparnamestr(jdxstringlist1.strings[li]) = 'XYPOINTS') or
        (jdxparnamestr(jdxstringlist1.strings[li]) = 'PEAKTABLE') or
        (jdxparnamestr(jdxstringlist1.strings[li]) = 'DATATABLE');
  if (jdxparnamestr(jdxstringlist1.strings[li]) = 'DATATABLE') then
    begin
      if pos('XYDATA',jdxstringlist1.strings[li]) > 0 then datatype := dtXYdata;
      if pos('XYPOINTS',jdxstringlist1.strings[li]) > 0 then datatype := dtXYpoints;
      if pos('PEAKTABLE',jdxstringlist1.strings[li]) > 0 then datatype := dtpeaktable;
    end;
  s       :=  1;
  maxt    := -10000;
  mint    :=  10000;
  maxXval := -10000;
  minXval :=  10000;
  case specsubtype of
    sstIRtransmittance : begin
                           offsetYval := 0;
                           offsetXval := 0;
                         end;
  end;  // case
  nvalues := 0;
  prevdif := false;  // important initialization!!
  if foundntuples then processntuples(page);
  if (jdxrec1.npoints = 0) then
    begin
      messagedlg('Could not determine NPOINTS from header (page '+inttostr(page)+')',mtError,[mbOK],0);
      specloaded := false;
      readjdx := false;
      exit;
    end;
  if (jdxrec1.npoints > maxpoints) then
    begin
      messagedlg('Too many data points! NPOINTS='+inttostr(jdxrec1.npoints),mtError,[mbOK],0);
      specloaded := false;
      readjdx := false;
      exit;
    end;
  repeat
    begin
      if (li < jdxstringlist1.count-1) then inc(li);
      tmpstr := jdxstringlist1.strings[li];
      if pos('##',tmpstr)= 0 then
        begin
          case datatype of
            dtXYdata    : outstr := parsejdxXYdataline(tmpstr);
            dtXYpoints  : outstr := parsejdxXYpointline(tmpstr);
            dtpeaktable : outstr := parsejdxXYpointline(tmpstr);
          end;
        end;
      if outstr <> '' then
        begin
          messagedlg('Data read error: '+outstr+
                    ' (line '+inttostr(li+1)+')',mtError,
                     [mbOK],0);
          specloaded := false;
          readjdx := false;
          exit;
        end;
    end;
  until (li = jdxstringlist1.count-1)
    or (pos('##END',jdxstringlist1.strings[li])=1)
    or ((pos('##PAGE=',jdxstringlist1.strings[li])=1) and foundntuples);
  if (li < jdxstringlist1.count-1) and
     (pos('##END',jdxstringlist1.strings[li])=1) and
     (pos('##END',jdxstringlist1.strings[li+1])=1) then inc(li);

  if li >= jdxstringlist1.count-2 then
    begin
      if form1.nextpage1.visible then
        begin
          li := -1;
          form1.nextpage1.caption := '&First page';
          page := 1;
        end;
    end else
    begin
      if form1.nextpage1.visible then form1.nextpage1.caption := '&Next page';
    end;
  if (spectype = stNMR) and (pos('HZ',jdxrec1.xunits)>0) then
    begin
      calcnewNMRparams;
      if (dppm = 0) and ((jdxrec1.firstx = 0) or (jdxrec1.lastx = 0)) then
        debugoutput('Hz to ppm conversion failed: no reference point!') else
        begin
          hz2ppm;
          updateNMRparams;
        end;
    end;
  if (spectype = stNMR) and (pos('PPM',jdxrec1.xunits)>0) and
     (SFO1 > 0) then showdeltaHz := true else showdeltaHz := false;
  maxYfound := maxt;
  minYfound := mint;
  XLo := xydata^[1].X;
  XHi := xydata^[jdxrec1.npoints].X;
  if (spectype = stMS) and (jdxrec1.peaktable <> '') then
    begin
      XLo := minXval - 10;
      if XLo < 10 then XLo := 10;
      XHi := maxXval + 20;
      if maxYfound > 100 then
        begin
          // readjust Y values to 100% base peak
          auxreal := 100/maxYfound;
          for i := 1 to jdxrec1.npoints do xydata^[i].Y := xydata^[i].Y*auxreal;
          maxYfound := maxYfound * auxreal;
          minYfound := minYfound * auxreal;
          jdxrec1.maxy := maxYfound;
          jdxrec1.miny := minYfound;
          jdxrec1.yfactor := 1;
        end;
    end;
  pdc := 1;
  adjustminmax;
  specloaded := true;
  LcursXval := XLo;
  LcursXindex := 1;
  LcursYval := xydata^[LcursXindex].Y;
  RcursXval := XHi;
  RcursXindex := jdxrec1.npoints;
  RcursYval := xydata^[RcursXindex].Y;
  form1.statusbar1.panels[0].text := inttostr(jdxrec1.npoints)+' points';
  readjdx := true;
end;

//=======================JCAMP-DX compression routines=====

function SQZencode(inval:longint):string;
var
  posSQZdig, negSQZdig : string[12];
  instr : string;
  char1 : char;
begin
  instr     := '';
  posSQZdig := '@ABCDEFGHI';  // 0-9
  negSQZdig :=  'abcdefghi';  // 1-9
  str(abs(inval):1,instr);
  char1 := instr[1];
  if inval < 0 then
    begin
      instr[1] := negSQZdig[pos(char1,'123456789')];
    end else
    begin
      instr[1] := posSQZdig[pos(char1,'0123456789')];
    end;
  SQZencode := instr;
end;

function DIFencode(difval:longint):string;
var
  posDIFdig, negDIFdig : string[12];
  instr : string;
  char1 : char;
begin
  instr     := '';
  posDIFdig := '%JKLMNOPQR';  // 0-9
  negDIFdig :=  'jklmnopqr';  // 1-9
  str(abs(difval):1,instr);
  char1 := instr[1];
  if difval < 0 then
    begin
      instr[1] := negDIFdig[pos(char1,'123456789')];
    end else
    begin
      instr[1] := posDIFdig[pos(char1,'0123456789')];
    end;
  DIFencode := instr;
end;

function DUPencode(dupcount:integer):string;
var
  DUPdig : string[12];
  instr  : string;
  char1  : char;
begin
  instr     := '';
  DUPdig := 'STUVWXYZs';  // 1-9
  str(dupcount:1,instr);
  char1 := instr[1];
  instr[1] := DUPdig[pos(char1,'123456789')];
  DUPencode := instr;
end;

procedure adjust_xdecimals(dtype:integer);
var
  abs_deltax : double;
begin
  case dtype of
    dtXYdata : begin
                 abs_deltax := abs((jdxrec1.firstx-jdxrec1.lastx)/(jdxrec1.npoints - 1));
                 xdecimals := 0;  // default
                 if abs_deltax < 10     then xdecimals := 1;
                 if abs_deltax < 1      then xdecimals := 2;
                 if abs_deltax < 0.1    then xdecimals := 3;
                 if abs_deltax < 0.01   then xdecimals := 4;
                 if abs_deltax < 0.001  then xdecimals := 5;
                 if abs_deltax < 0.0001 then xdecimals := 6;
               end;
  end;  // case
end;

procedure writeXYdata;
var
  i : integer;
  ystr, wstr, auxstr : string;
  auxfloat : extended;
  difdupline : string;
  actY,prevY : longint;
  actdif,prevdif : longint;
  dupcount : integer;
  items : integer;
begin
  adjust_xdecimals(dtXYdata);
  if datamode = dmAFFN then
    begin
      wstr := '';
      for i := 1 to jdxrec1.npoints do
        begin
          if (i mod 6) = 1 then
            begin
              if wstr <> '' then jdxstringlist2.add(wstr);
              str(xydata^[i].x:4:xdecimals,wstr);
            end;
          auxfloat := xydata^[i].y/jdxrec1.yfactor;
          str(round(auxfloat):1,ystr);
          wstr := wstr + ' ' + ystr;
        end;
      if wstr <> '' then jdxstringlist2.add(wstr);
    end else
    begin  // datamode = dmASDF (DIFDUP compression)
      difdupline := '';
      items := 0;
      // prepare first line (this is a special case)
      i := 1;
      str(xydata^[i].x:4:xdecimals,difdupline); // start new line with X value
      auxfloat := xydata^[i].y/jdxrec1.yfactor;
      actY := round(auxfloat);
      auxstr := SQZencode(actY);  // first Y value ==> always SQZ
      difdupline := difdupline + auxstr;
      prevY := actY;
      dupcount := 1;
      items := 1;   // one Y item is already present
      repeat    // ....until i = npoints
        begin
          inc(i);
          if length(difdupline) > 70 then   // line is full ==> write to disk
            begin
              if dupcount > 1 then  // ==> make sure nothing gets lost
                begin
                  auxstr := DUPencode(dupcount);
                  difdupline := difdupline + auxstr;
                  dupcount := 1;
                end;
              jdxstringlist2.add(difdupline); // write line
              difdupline := '';               // clear line
              dec(i);                         // jump back one value
              str(xydata^[i].x:4:xdecimals,difdupline); // start new line with X value
              auxfloat := xydata^[i].y/jdxrec1.yfactor;
              actY := round(auxfloat);
              auxstr := SQZencode(actY);  // first Y value ==> always SQZ
              difdupline := difdupline + auxstr;
              prevY := actY;
              dupcount := 1;
              items := 1;   // one Y item is already present
            end else
            begin
              inc(items);
              if items = 2 then
                begin  // must be a difference in any case
                  auxfloat := xydata^[i].y/jdxrec1.yfactor;
                  actY := round(auxfloat);
                  actdif := actY - prevY;
                  auxstr := DIFencode(actdif);
                  difdupline := difdupline + auxstr;
                  prevY := actY;
                  prevdif := actdif;
                end else
                begin  // can be difference or duplicate diff.
                  auxfloat := xydata^[i].y/jdxrec1.yfactor;
                  actY := round(auxfloat);
                  actdif := actY - prevY;
                  if (actdif = prevdif) then
                    begin
                      inc(dupcount);
                      prevY := actY;
                    end else
                    begin
                      if (dupcount > 1) then
                        begin  // write accumulated duplicates in any case
                          auxstr := DUPencode(dupcount);
                          difdupline := difdupline + auxstr;
                          dupcount := 1;
                        end;
                      auxstr := DIFencode(actdif);
                      difdupline := difdupline + auxstr;
                      prevY := actY;
                      prevdif := actdif;
                    end;
                  if (i = jdxrec1.npoints) then
                    begin                   // end of array
                      if dupcount > 1 then  // ==> make sure nothing gets lost
                        begin
                          auxstr := DUPencode(dupcount);
                          difdupline := difdupline + auxstr;
                          dupcount := 1;
                        end;
                    end;
                end;
            end;
        end;
      until i = jdxrec1.npoints;
      if difdupline <> '' then jdxstringlist2.add(difdupline);
    end;
end;

procedure writeXYpoints;
var
  i : integer;
  xstr, ystr : string;
begin
  for i := 1 to jdxrec1.npoints do
    begin
      str(xydata^[i].X:1:3,xstr);
      str(xydata^[i].Y:1:5,ystr);
      jdxstringlist2.Add(xstr+', '+ystr);
    end;
end;

procedure writepeaktable;
var
  i : integer;
  xstr, ystr : string;
begin
  for i := 1 to jdxrec1.npoints do
    begin
      str(xydata^[i].X:1:1,xstr);
      str(xydata^[i].Y:1:2,ystr);
      jdxstringlist2.Add(xstr+', '+ystr);
    end;
end;

procedure writejdxfile(jdxfilename:string);
var
  auxstr   : string;
begin
  jdxstringlist2.clear;
  with jdxrec1 do
    begin
      jdxstringlist2.Add('##TITLE= '+title);
      jdxstringlist2.Add('##JCAMP-DX= 4.24');
      jdxstringlist2.Add('##DATA TYPE= '+spectype);
      jdxstringlist2.Add('##ORIGIN= '+origin);
      jdxstringlist2.Add('##OWNER= '+owner);
      jdxstringlist2.Add('##DATE= '+date);
      jdxstringlist2.Add('##TIME= '+time);
      jdxstringlist2.Add('##SPECTROMETER/DATA SYSTEM= '+spectrometer);
      if (jdxUnit1.spectype = stNMR) then
        begin
          if SFO1 > 0 then
            begin
              str(SFO1:1:4,auxstr);
              jdxstringlist2.Add('##.OBSERVE FREQUENCY= '+auxstr);
            end;
          if NMRnucleus <> '' then
            jdxstringlist2.Add('##.OBSERVE NUCLEUS= '+NMRnucleus);
          if NMRscans > 0 then
            begin
              str(NMRscans,auxstr);
              jdxstringlist2.Add('##.AVERAGES= '+auxstr);
            end;
        end;
      jdxstringlist2.Add('##DATA PROCESSING= '+params);
      jdxstringlist2.Add('##XUNITS= '+xunits);
      jdxstringlist2.Add('##YUNITS= '+yunits);
      str(xfactor:1:8,auxstr);
      jdxstringlist2.Add('##XFACTOR= '+auxstr);
      str(yfactor:1:8,auxstr);
      jdxstringlist2.Add('##YFACTOR= '+auxstr);
      str(firstx:1:6,auxstr);
      jdxstringlist2.Add('##FIRSTX= '+auxstr);
      str(lastx:1:6,auxstr);
      jdxstringlist2.Add('##LASTX= '+auxstr);
      jdxstringlist2.Add('##NPOINTS= '+inttostr(npoints));
    end;
  str(xydata^[1].Y:1:6,auxstr);
  jdxstringlist2.Add('##FIRSTY= '+auxstr);
  str(maxYfound:1:6,auxstr);
  jdxstringlist2.Add('##MAXY= '+auxstr);
  str(minYfound:1:6,auxstr);
  jdxstringlist2.Add('##MINY= '+auxstr);
  case datatype of
    dtXYdata    :  begin
                     jdxstringlist2.Add('##XYDATA= (X++(Y..Y))');
                     writexydata;
                   end;
    dtXYpoints  :  begin
                     jdxstringlist2.Add('##XYPOINTS= (XY..XY)');
                     writexypoints;
                   end;
    dtpeaktable :  begin
                     jdxstringlist2.Add('##PEAK TABLE= (XY..XY)');
                     writepeaktable;
                   end;
  end;  // case
  jdxstringlist2.Add('##END= ');
  try
    jdxstringlist2.savetofile(jdxfilename);
  except
    messagedlg('Could not save spectrum to file '+jdxfilename, mtError,[mbOK],0);
    exit;
  end;
end;

procedure readjdxfile(jdxfilename:string);
var
  i : longint;
  endcount : integer;
  pagecount : integer;
begin
  specloaded := false;
  jdxstringlist1.Clear;
  foundntuples := false;
  page := 1;
  try
    jdxstringlist1.loadfromfile(jdxfilename);
  except
    messagedlg('Could not load spectrum from file '+jdxfilename, mtError,[mbOK],0);
    exit;
  end;
    if jdxstringlist1.Count > 0 then
  begin
    for i := 1 to jdxstringlist1.count-1 do
      jdxstringlist1.strings[i] := trimleft(jdxstringlist1.strings[i]);
    initjdxrec;
    initNMRparams;
    datatype := dtXYdata;
    li := -1;
    readjdx;
    endcount := 0;
    pagecount := 0;
    for i := 1 to jdxstringlist1.count-1 do
      begin
        if pos('##END=',jdxstringlist1.strings[i])=1 then inc(endcount);
        if pos('##PAGE=',jdxstringlist1.strings[i])=1 then inc(pagecount);
      end;
    if (endcount > 1) or (pagecount > 1) then form1.nextpage1.visible := true else
                         form1.nextpage1.visible := false;
  end;
end;

procedure dumpXYvalues(dumpfilename:string);
var
//  dumpfile : text;
  i : integer;
  xstr, ystr : string;
begin
  if not specloaded then exit;
(*
  assignfile(dumpfile,dumpfilename);
  rewrite(dumpfile);
  for i := 1 to jdxrec1.npoints do
    begin
      writeln(dumpfile,xydata^[i].X:1:5,TAB,xydata^[i].Y:1:5);
    end;
  closefile(dumpfile);
*)
  with jdxstringlist2 do
    begin
      clear;
      for i := 1 to jdxrec1.npoints do
        begin
          str(xydata^[i].X:1:5,xstr);
          str(xydata^[i].Y:1:5,ystr);
          add(xstr+TAB+ystr);
        end;
      try
        savetofile(dumpfilename);
      except
        messagedlg('Could not write file!',mtError,[mbOK],0);
      end;
    end;
end;

function findXindex(Xval:double):integer;
var
  i       : integer;
  auxint  : integer;
  imem    : integer;
  mindiff : double;
begin
  if ((iniXlo-iniXHi)=0) or ((jdxrec1.npoints-1)=0) then auxint := -1 else
    begin
      if datatype = dtXYdata then
        // a) routine for equally spaced XY data
        auxint := round(((iniXLo-Xval)/(iniXLo-iniXHi))*(jdxrec1.npoints-1) + 1   ) else
        begin
        // b) routine for non-equally spaced XY data
          mindiff := abs(Xlo-XHi);
          imem := 1;
          for i := 1 to jdxrec1.npoints do
            begin
              if abs(Xval-xydata^[i].X) < mindiff then
                begin
                  mindiff := abs(Xval-xydata^[i].X);
                  imem := i;
                end;
            end;
          if mindiff < abs(XLo-XHi)*0.002 then auxint := imem else auxint := 0;
        end;
      if auxint > jdxrec1.npoints then auxint := jdxrec1.npoints;
    end;
  findXindex := auxint;  // error is signalled as -1
end;

procedure clipbyval(loXval,hiXval:double);
var
  i, j, orignpoints : longint;
  auxfloat : double;
  minXtmp, maxXtmp : double;
  minYtmp, maxYtmp : double;
  ecount : longint;
begin
  if loXval > hiXval then
    begin
      auxfloat := loXval;
      loXval := hiXval;
      hiXval := auxfloat;
    end;
  if (loXval > maxXval) or (hiXval < minXval) then exit;
  if loXval < minXval then loXval := minXval;
  if hiXval > maxXval then hiXval := maxXval;
  if (loXval = minXval) and (hiXval = maxXval) then exit else
    begin
      orignpoints := jdxrec1.npoints;
      minYtmp := jdxrec1.maxY;
      maxYtmp := jdxrec1.minY;
      minXtmp := maxXval;
      maxXtmp := minXval;
      ecount := 0;
      for i := 1 to orignpoints do
        begin
          auxfloat := xydata^[i].X;
          while ((auxfloat < loXval) or (auxfloat > hiXval)) and (jdxrec1.npoints > i) do
            begin
              for j := i to (orignpoints-1) do
                begin
                  xydata^[j].X := xydata^[j+1].X;
                  xydata^[j].Y := xydata^[j+1].Y;
                end;
              dec(jdxrec1.npoints);
              inc(ecount);
              auxfloat := xydata^[i].X;
            end;
        end;
      if ((auxfloat < loXval) or (auxfloat > hiXval)) then dec(jdxrec1.npoints);
      for i := 1 to jdxrec1.npoints do
        begin
          if xydata^[i].X < minXtmp then minXtmp := xydata^[i].X;
          if xydata^[i].X > maxXtmp then maxXtmp := xydata^[i].X;
          if xydata^[i].Y < minYtmp then minYtmp := xydata^[i].Y;
          if xydata^[i].Y > maxYtmp then maxYtmp := xydata^[i].Y;
        end;
      jdxrec1.firstX := xydata^[1].X;
      jdxrec1.lastX  := xydata^[jdxrec1.npoints].X;
      jdxrec1.firstY := xydata^[1].Y;
      jdxrec1.maxY   := maxYtmp;
      jdxrec1.minY   := minYtmp;
      maxXval        := maxXtmp;
      minXval        := minXtmp;
      XLo := xydata^[1].X;
      XHi := xydata^[jdxrec1.npoints].X;
      if (spectype = stMS) and (jdxrec1.peaktable <> '') then
        begin
          XLo := minXtmp - 10;
          if XLo < 10 then XLo := 10;
          XHi := maxXtmp + 20;
        end;
      iniXLo := Xlo;
      iniXHi := XHi;
      LcursXval  := XLo;
      RcursXval  := XHi;
      XindexLo := 1;
      XindexHi := jdxrec1.npoints;
      LCursxindex := XindexLo;
      RcursXindex := XindexHi;
      form1.statusbar1.panels[0].text := inttostr(jdxrec1.npoints)+' points';
    end;
end;

procedure clipbyindex(loXindex,hiXindex:longint);
var
  i, orignpoints : longint;
  auxint : longint;
  auxfloat : double;
  minXtmp, maxXtmp : double;
  minYtmp, maxYtmp : double;
  ecount : integer;
begin
  if loXindex > hiXindex then
    begin
      auxint   := loXindex;
      loXindex := hiXindex;
      hiXindex := auxint;
    end;
  if (loXindex > jdxrec1.npoints) or (hiXindex < 1) then exit;
  if loXindex < 1 then loXindex := 1;
  if hiXindex > jdxrec1.npoints then hiXindex := jdxrec1.npoints;
  if (loXindex = 1) and (hiXindex = jdxrec1.npoints) then exit else
    begin
      orignpoints := jdxrec1.npoints;
      minYtmp := jdxrec1.maxY;
      maxYtmp := jdxrec1.minY;
      minXtmp := maxXval;
      maxXtmp := minXval;
      for i := loXindex to hiXindex do
        begin
          xydata^[i-(loXindex-1)].X := xydata^[i].X;
          if maxXtmp < xydata^[i].X then maxXtmp := xydata^[i].X;
          if minXtmp > xydata^[i].X then minXtmp := xydata^[i].X;
          auxfloat := xydata^[i].Y;
          xydata^[i-(loXindex-1)].Y := auxfloat;
          if auxfloat < minYtmp then minYtmp := auxfloat;
          if auxfloat > maxYtmp then maxYtmp := auxfloat;
        end;
      jdxrec1.npoints := (hiXindex - loXindex) + 1;
      jdxrec1.firstX := xydata^[1].X;
      jdxrec1.lastX  := xydata^[jdxrec1.npoints].X;
      jdxrec1.firstY := xydata^[1].Y;
      jdxrec1.maxY   := maxYtmp;
      jdxrec1.minY   := minYtmp;
      maxXval        := maxXtmp;
      minXval        := minXtmp;
      XLo := xydata^[1].X;
      XHi := xydata^[jdxrec1.npoints].X;
      iniXLo := Xlo;
      iniXHi := XHi;
      LcursXval  := XLo;
      RcursXval  := XHi;
      XindexLo := 1;
      XindexHi := jdxrec1.npoints;
      LCursxindex := XindexLo;
      RcursXindex := XindexHi;
      form1.statusbar1.panels[0].text := inttostr(jdxrec1.npoints)+' points';
    end;
end;

procedure writepltfile(pltfilename:string);
const
  PltBaseX : integer = 1000;
  PltFullX : integer = 9000;
  PltBaseY : integer = 1000;
  PltFullY : integer = 6000;
  OutRes   : integer = 2;
var
  PrevX, PrevY, PixX, PixY, auxint, i  : integer;
  aux1 : real;
  auxreal : real;
  outstr  : string;
  ScanCM  : real;
  s : integer;
  xindex, startxindex, endxindex : integer;
  Yaccu : extended;
  xstep, nval : integer;
  Yscaleval : double;
  Yscalecorr : integer;
  Xscaleval : double;
  xsdc : integer;
  startXval : double;
  endXval   : double;
  frame : trect;
  wstr, xstr, ystr : string;
begin
  if not specloaded then exit;
  with jdxstringlist2 do
    begin
      clear;
      // Plotter initialisation
      wstr := 'IN;CS0;CA7;TL0,1.0;VS10;SP1;';
      add(wstr);
      wstr := '';
      xstep := ((XindexHi-XindexLo) + 1) div (6*fullX);  // screen: 4*fullX
      if xstep < 1 then xstep := 1;
      if xstep > 4 then xstep := 8;
      if pdc = 1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexLo;
              endxindex   := XindexHi;
            end else
            begin
              startxindex := 1;
              endxindex   := jdxrec1.npoints;
            end;
          startXval   := XLo;
          endXval     := XHi;
        end;
      if pdc = -1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexHi;
              endxindex   := XindexLo;
            end else
            begin
              startxindex := jdxrec1.npoints;
              endxindex   := 1;
            end;
          startXval   := XHi;
          endXval     := XLo;
        end;
      // Grid
      if grid then
        begin
          add('LT 1;');
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          wstr := '';
          for i := YscaleLo to YscaleHi do
            begin
              pixY := pltbaseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*pltfullY );
              if (pixY <= (pltbaseY+pltfullY)) and (pixY >= pltbaseY) then
                begin
                  str(pltbaseX:1,xstr);
                  str(pixY,ystr);
                  wstr := wstr + 'PU '+xstr+','+ystr+';';
                  str((pltbaseX+pltfullX):1,xstr);
                  wstr := wstr + 'PD '+xstr+','+ystr+';';
                  add(wstr);
                  wstr := '';
                end;
            end;
          // vertical lines
          if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
          i := XscaleLo;
          while i <> XscaleHi do
            begin
              pixX := pltbaseX + abs(round(pltfullX * ((startXval - (i/Xscalefactor)) /
                      (abs(XHi - XLo) ) )));
              str(pixX:1,xstr);
              str((pltbaseY):1,ystr);
              wstr := wstr + 'PU '+xstr+','+ystr+';';
              str((pltbaseY+pltfullY):1,ystr);
              wstr := wstr + 'PD '+xstr+','+ystr+';';
              add(wstr);
              wstr := '';
              i := i + xsdc;
            end;
          add('LT;');
        end;  // end of grid routine
      if datatype = dtXYdata then  // equally spaced X values
        begin
          prevX := pltbaseX;
          prevY := pltbaseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*pltfullY );
          nval  := 0;
          Yaccu := 0;
          str(prevX:1,xstr);
          str(prevY:1,ystr);
          wstr := 'PU '+xstr+','+ystr+';';
          add(wstr);
          wstr := '';
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := pltbaseX + round( (abs(startxindex-xindex)/abs(XindexHi-XindexLo)) *pltfullX );
                  pixY := pltbaseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*pltfullY );
                  str(pixX:1,xstr);
                  str(pixY:1,ystr);
                  wstr := wstr+'PD '+xstr+','+ystr+';';
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
        end;
      if datatype = dtXYpoints then  // non-equally-spaced X values
        begin
          prevX := pltbaseX;
          prevY := pltbaseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*pltfullY );
          nval  := 0;
          Yaccu := 0;
          str(prevX:1,xstr);
          str(prevY:1,ystr);
          wstr := 'PU '+xstr+','+ystr+';';
          add(wstr);
          wstr := '';
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := pltbaseX + abs(round(pltfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
                  pixY := pltbaseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*pltfullY );
                  str(pixX:1,xstr);
                  str(pixY:1,ystr);
                  wstr := wstr + 'PD '+xstr+','+ystr+';';
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
        end;
      if datatype = dtpeaktable then    // MS peak list etc.
        begin
          xindex := startxindex - pdc;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              pixX := pltbaseX + abs(round(pltfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
              pixY := pltbaseY+round( ((xydata^[xindex].Y-minYadj)/((maxyadj-minYadj)))*pltfullY );
              if ((xydata^[xindex].X > XLo) and (xydata^[xindex].X < XHi)) or
                 ((xydata^[xindex].X < XLo) and (xydata^[xindex].X > XHi)) then
                begin
                  str(pixX:1,xstr);
                  str(pltbaseY:1,ystr);
                  wstr := wstr + 'PU '+xstr+','+ystr+';';
                  str(pixY:1,ystr);
                  wstr := wstr + 'PD '+xstr+','+ystr+';';
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
        end;
      // frame, X scale, Y scale, comment etc.
      str(pltbaseX:1,xstr);
      str(pltbaseY:1,ystr);
      wstr := 'PU '+xstr+','+ystr+';';
      str((pltbaseX+pltfullX):1,xstr);
      wstr := wstr + 'PD '+xstr+','+ystr+';';
      str((pltbaseY+pltfullY):1,ystr);
      wstr := wstr + 'PD '+xstr+','+ystr+';';
      str(pltbaseX:1,xstr);
      wstr := wstr + 'PD '+xstr+','+ystr+';';
      str(pltbaseY:1,ystr);
      wstr := wstr + 'PD '+xstr+','+ystr+';';
      add(wstr);
      wstr := '';
      // Y scale
      if spectype <> stNMR then
        begin
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          pixX := pltbaseX - 50;
          for i := YscaleLo to YscaleHi do
            begin
              pixY := pltbaseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*pltfullY );
              if (pixY <= (pltbaseY+pltfullY)) and (pixY >= pltbaseY) then
                begin
                  str(pixX:1,xstr);
                  str(pixY,ystr);
                  wstr := wstr + 'PU '+xstr+','+ystr+';';
                  str(pltbaseX:1,xstr);
                  wstr := wstr + 'PD '+xstr+','+ystr+';';
                  Yscaleval := (i*Yscalecorr)/Yscalefactor;
                  str(Yscaleval:1:Yscaledigits,outstr);
                  while length(outstr) < 5 do outstr := ' '+outstr;
                  str((pltbaseX-600):1,xstr);
                  str((pixY-60):1,ystr);
                  wstr := wstr+'PU '+xstr+','+ystr+';LB'+OutStr+ETX;
                  add(wstr);
                  wstr := '';
                  //textout(baseX-(10+textwidth(outstr)),(pixY-textheight(outstr) div 2),outstr);
                end;
            end;
          //textout(10, ((baseY+fullY) div 2) - (textheight(jdxrec1.yunits) div 2),jdxrec1.yunits);
          str((pltbaseX-600):1,xstr);
          str((pltbaseY+pltfullY+200):1,ystr);
          outstr := jdxrec1.yunits;
          wstr := wstr + 'PU '+xstr+','+ystr+';LB'+outstr+ETX;
          add(wstr);
          wstr := '';
        end;
      // X scale
      pixY := pltbaseY - 50;
      if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
      if abs(XscaleLo-XscaleHi) >= 100 then Xscalemod := 50 else
        if abs(XscaleLo-XscaleHi) >= 20 then Xscalemod := 10 else
          if abs(XscaleLo-XscaleHi) >= 5 then Xscalemod := 2 else Xscalemod := 1;
      Xscalemod := Xscalemod div round(fullX/600);
      if Xscalemod < 1 then Xscalemod := 1;
      i := XscaleLo;
      if wstr <> '' then
        begin
          add(wstr);
          wstr := '';
        end;
      while i <> XscaleHi do
        begin
          pixX := pltbaseX + abs(round(pltfullX * ((startXval - (i/Xscalefactor)) /
                  (abs(XHi - XLo) ) )));
          if ((i mod Xscalemod) = 0) then
            begin
              str(pixX:1,xstr);
              str((pixY-50):1,ystr);
              wstr := wstr + 'PU '+xstr+','+ystr+';';
            end else
            begin
              str(pixX:1,xstr);
              str(pixY:1,ystr);
              wstr := wstr + 'PU '+xstr+','+ystr+';';
            end;
          str(pltbaseY:1,ystr);
          wstr := wstr + 'PD '+xstr+','+ystr+';';
          Xscaleval := (i/Xscalefactor);
          str(Xscaleval:1:Xscaledigits,outstr);
          while length(outstr) < 5 do outstr := ' '+outstr;
          if ((i mod Xscalemod) = 0) then
            begin
              str(round(pixX-50*Length(outstr)):1,xstr);
              str((pixY-250):1,ystr);
              wstr := wstr + 'PU '+xstr+','+ystr+';LB'+outstr+ETX;
              add(wstr);
              wstr := '';
            end;
          i := i + xsdc;
        end;
      outstr := jdxrec1.xunits;
      str(((pltbaseX+pltfullX)-100*length(outstr)):1,xstr);
      str((pltbaseY-600):1,ystr);
      wstr := wstr + 'PU '+xstr+','+ystr+';LB'+outstr+ETX;
      add(wstr);
      wstr := '';
      // end of X/Y scale routine
      outstr := jdxrec1.title;
      if outstr <> '' then
        begin
          str(((pltbaseX+(pltfullX div 2))-50*length(outstr)):1,xstr);
          str((pltbaseY+pltfullY+200):1,ystr);
          wstr := 'PU '+xstr+','+ystr+';LB'+outstr+ETX;
          add(wstr);
          wstr := '';
        end;
      try
        savetofile(pltfilename);
      except
        messagedlg('Could not write plot file!',mtError,[mbOK],0);
      end;
    end;
end;

procedure writesvgfile(svgfilename:string);
const
  SVGwidth  : integer = 1200;
  SVGheight : integer = 600;
  PltBaseX : integer = 1000;
  PltFullX : integer = 9000;
  PltBaseY : integer = 1000;
  PltFullY : integer = 6000;
  OutRes   : integer = 2;
var
  PrevX, PrevY, PixX, PixY, auxint, i  : integer;
  aux1 : real;
  auxreal : real;
  outstr  : string;
  ScanCM  : real;
  s : integer;
  xindex, startxindex, endxindex : integer;
  Yaccu : extended;
  xstep, nval : integer;
  Yscaleval : double;
  Yscalecorr : integer;
  Xscaleval : double;
  xsdc : integer;
  startXval : double;
  endXval   : double;
  frame : trect;
  wstr, xstr, ystr : string;
  svgXleft, svgXright, svgYtop, svgYbottom : integer;
  svgbaseX, svgfullX, svgbaseY, svgfullY : integer;
  OffsetLeft,OffsetBottom : integer;
  lblX, lblY : integer;
  lblstr     : string;
begin
  if not specloaded then exit;
  svgXleft     := 0;
  svgXright    := SVGwidth;
  svgYtop      := 0;
  svgYbottom   := SVGheight;
  svgbaseX     := 100;
  svgbaseY     := SVGheight - 60;
  svgfullX     := SVGwidth-(svgbaseX+20);
  svgfullY     := -(svgbaseY-35);  // must be a negative number!
  with jdxstringlist2 do
    begin
      clear;
      // write SVG file header
      add('<?xml version="1.0" standalone="no"?>');
      add('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20010904//EN"');
      add('"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">');
      add('<svg width="16cm" height="8cm" viewBox="0 0 '+inttostr(SVGwidth)+' '+inttostr(SVGheight)+'"');
      add('xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">');
      add('<desc>Spectrum of '+jdxrec1.title+'</desc>');
      add('<!-- generated by '+appname+' '+appversion+' (N. Haider, 2003, 2004) -->');
      wstr := '';
      xstep := ((XindexHi-XindexLo) + 1) div (4*fullX);  // screen: 4*fullX
      if xstep < 1 then xstep := 1;
      if xstep > 4 then xstep := 8;
      if pdc = 1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexLo;
              endxindex   := XindexHi;
            end else
            begin
              startxindex := 1;
              endxindex   := jdxrec1.npoints;
            end;
          startXval   := XLo;
          endXval     := XHi;
        end;
      if pdc = -1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexHi;
              endxindex   := XindexLo;
            end else
            begin
              startxindex := jdxrec1.npoints;
              endxindex   := 1;
            end;
          startXval   := XHi;
          endXval     := XLo;
        end;
      // Grid
      if grid then
        begin
          add('<g stroke="blue" stroke-width="1">');
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          wstr := '';
          for i := YscaleLo to YscaleHi do
            begin
              pixY := svgbaseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*svgfullY );
              if (pixY >= (svgbaseY+svgfullY)) and (pixY <= svgbaseY) then
                begin
                  //moveto(pixX,pixY);
                  //lineto(baseX,pixY);
                  add('<line x1="'+inttostr(svgbaseX)+'" y1="'+inttostr(PixY)+'" x2="'+inttostr(svgbaseX+svgfullX)+'" y2="'+inttostr(PixY)+'" />');
                end;
            end;
          // vertical lines
          if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
          i := XscaleLo;
          while i <> XscaleHi do
            begin
              pixX := svgbaseX + abs(round(svgfullX * ((startXval - (i/Xscalefactor)) /
                      (abs(XHi - XLo) ) )));
              //moveto(pixX,pixY);
              add('<line x1="'+inttostr(PixX)+'" y1="'+inttostr(svgbaseY)+'" x2="'+inttostr(PixX)+'" y2="'+inttostr(svgbaseY+svgfullY)+'" />');
              i := i + xsdc;
            end;
          add('</g>');
        end;  // end of grid routine
      if datatype = dtXYdata then  // equally spaced X values
        begin
          add('<polyline fill="none" stroke="black" stroke-width="2"');
          add('points="');
          prevX := svgbaseX;
          prevY := svgbaseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*svgfullY );
          nval  := 0;
          Yaccu := 0;
          str(prevX:1,xstr);
          str(prevY:1,ystr);
          wstr := xstr+','+ystr+' ';
          //moveto(prevX,prevY);
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := svgbaseX + round( (abs(startxindex-xindex)/abs(XindexHi-XindexLo)) *svgfullX );
                  pixY := svgbaseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*svgfullY );
                  //lineto(pixX,pixY);
                  if (pixX <> prevX) or (pixY <> prevY) then
                    begin
                      str(pixX:1,xstr);
                      str(pixY:1,ystr);
                      wstr := wstr+xstr+','+ystr+' ';
                    end;
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('" />');
        end;
      if datatype = dtXYpoints then  // non-equally-spaced X values
        begin
          add('<polyline fill="none" stroke="black" stroke-width="2"');
          add('points="');
          prevX := svgbaseX;
          prevY := svgbaseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*svgfullY );
          nval  := 0;
          Yaccu := 0;
          //moveto(prevX,prevY);
          str(prevX:1,xstr);
          str(prevY:1,ystr);
          wstr := xstr+','+ystr+' ';
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := svgbaseX + abs(round(svgfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
                  pixY := svgbaseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*svgfullY );
                  //lineto(pixX,pixY);
                  if (pixX <> prevX) or (pixY <> prevY) then
                    begin
                      str(pixX:1,xstr);
                      str(pixY:1,ystr);
                      wstr := wstr + xstr+','+ystr+' ';
                    end;
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('" />');
        end;
      if datatype = dtpeaktable then    // MS peak list etc.
        begin
          xindex := startxindex - pdc;
          add('<g stroke="black" stroke-width="1">');
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              pixX := svgbaseX + abs(round(svgfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
              pixY := svgbaseY+round( ((xydata^[xindex].Y-minYadj)/((maxyadj-minYadj)))*svgfullY );
              if ((xydata^[xindex].X > XLo) and (xydata^[xindex].X < XHi)) or
                 ((xydata^[xindex].X < XLo) and (xydata^[xindex].X > XHi)) then
                begin
                  //moveto(pixX,baseY);
                  //lineto(pixX,pixY);
                  str(pixX:1,xstr);
                  str(svgbaseY:1,ystr);
                  wstr := '<line x1="'+xstr+'" y1="'+ystr+'" ';
                  str(pixY:1,ystr);
                  wstr := wstr + 'x2="'+xstr+'" y2="'+ystr+'" />';
                  add(wstr);
                  wstr := '';
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('</g>');
        end;
      // frame, X scale, Y scale, comment etc.
      //moveto(baseX,baseY);
      //lineto(baseX+fullX,baseY);
      //lineto(baseX+fullX,baseY+fullY);
      //lineto(baseX,baseY+fullY);
      //lineto(baseX,baseY);
      // frame around spectrum
      add('<rect x="'+inttostr(svgbaseX)+'" y="'+inttostr((svgbaseY+svgfullY))+'" width="'+inttostr(svgfullX)+'" height="'+inttostr(-svgfullY)+'"');
      add('fill="none" stroke="black" stroke-width="2"/>');
      // Y scale
      add('<g stroke="black" stroke-width="1">');
      if spectype <> stNMR then
        begin
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          pixX := svgbaseX - 10;
          for i := YscaleLo to YscaleHi do
            begin
              pixY := svgbaseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*svgfullY );
              if (pixY >= (svgbaseY+svgfullY)) and (pixY <= svgbaseY) then
                begin
                  //moveto(pixX,pixY);
                  //lineto(baseX,pixY);
                  Yscaleval := (i*Yscalecorr)/Yscalefactor;
                  str(Yscaleval:1:Yscaledigits,outstr);
                  add('<line x1="'+inttostr(svgbaseX-10)+'" y1="'+inttostr(PixY)+'" x2="'+inttostr(svgbaseX)+'" y2="'+inttostr(PixY)+'" />');
                  add('<text x="'+inttostr(svgbaseX-15)+'" y="'+inttostr(PixY+4)+'" font-family="Verdana" font-size="14"');
                  add('fill="black" text-anchor="end">'+OutStr+'</text>');
                  //textout(baseX-(10+textwidth(outstr)),(pixY-textheight(outstr) div 2),outstr);
                end;
            end;
          //textout(10, ((baseY+fullY) div 2) - (textheight(jdxrec1.yunits) div 2),jdxrec1.yunits);
          outstr := jdxrec1.yunits;
          add('<text x="'+inttostr(svgbaseX-60)+'" y="'+inttostr(svgbaseY+svgfullY-20)+'" font-family="Verdana" font-size="14"');
          add('fill="black" text-anchor="start">'+OutStr+'</text>');
          wstr := '';
        end;
      // X scale
      pixY := svgbaseY + 10;
      if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
      if abs(XscaleLo-XscaleHi) >= 100 then Xscalemod := 50 else
        if abs(XscaleLo-XscaleHi) >= 20 then Xscalemod := 10 else
          if abs(XscaleLo-XscaleHi) >= 5 then Xscalemod := 2 else Xscalemod := 1;
      Xscalemod := Xscalemod div round(fullX/600);  // should be derived from screen-fullX
      if Xscalemod < 1 then Xscalemod := 1;
      i := XscaleLo;
      while i <> XscaleHi do
        begin
          pixX := svgbaseX + abs(round(svgfullX * ((startXval - (i/Xscalefactor)) /
                  (abs(XHi - XLo) ) )));
          if ((i mod Xscalemod) = 0) then
            begin
              //moveto(pixX,pixY+5);
              Xscaleval := (i/Xscalefactor);
              str(Xscaleval:1:Xscaledigits,outstr);
              add('<text x="'+inttostr(PixX)+'" y="'+inttostr(svgbaseY+30)+'" font-family="Verdana" font-size="14"');
              add('fill="black" text-anchor="middle">'+OutStr+'</text>');
              add('<line x1="'+inttostr(PixX)+'" y1="'+inttostr(svgbaseY)+'" x2="'+inttostr(PixX)+'" y2="'+inttostr(svgbaseY+15)+'" />');
            end else
            begin
              //moveto(pixX,pixY);
              add('<line x1="'+inttostr(PixX)+'" y1="'+inttostr(svgbaseY)+'" x2="'+inttostr(PixX)+'" y2="'+inttostr(svgbaseY+10)+'" />');
            end;
          i := i + xsdc;
        end;
      //textout((baseX+fullX) - textwidth(jdxrec1.xunits), baseY+((paintbox1.height-baseY) div 2)-(textheight(jdxrec1.xunits) div 2),jdxrec1.xunits);
      outstr := jdxrec1.xunits;
      add('<text x="'+inttostr(svgbaseX+svgfullX)+'" y="'+inttostr(svgbaseY+45)+'" font-family="Verdana" font-size="14"');
      add('fill="black" text-anchor="end">'+OutStr+'</text>');
      // end of X/Y scale routine
      outstr := jdxrec1.title;
      if outstr <> '' then
        begin
          add('<text x="'+inttostr(svgbaseX+(svgfullX div 2))+'" y="'+inttostr(svgbaseY+svgfullY-20)+'" font-family="Verdana" font-size="14"');
          add('fill="black" text-anchor="middle">'+OutStr+'</text>');
        end;
      add('</g>');
      add('</svg>');
      try
        savetofile(svgfilename);
      except
        messagedlg('Could not write SVG file!',mtError,[mbOK],0);
      end;
    end;
end;

procedure writepsfile(psfilename:string);
const
  PSBaseX : double  = 57;
  PSFullX : double  = 440;
  PSBaseY : double  = 47;
  PSFullY : double  = 283;
  OutRes  : integer = 2;
  cbo     : char    = #123;
  cbc     : char    = #125;
  CtrlD   = #4;
var
  auxint, i  : integer;
  PrevPSX, PrevPSY, PSX, PSY  : double;
  aux1 : real;
  auxreal : real;
  outstr  : string;
  s : integer;
  xindex, startxindex, endxindex : integer;
  Yaccu : extended;
  xstep, nval : integer;
  Yscaleval : double;
  Yscalecorr : integer;
  Xscaleval : double;
  xsdc : integer;
  startXval : double;
  endXval   : double;
  wstr, xstr, ystr : string;
  PSXleft, PSXright, PSYtop, PSYbottom : integer;
begin
  if not specloaded then exit;
  with jdxstringlist2 do
    begin
      clear;
      // write EPS file header
      add('%!PS-Adobe-2.0 EPSF-1.2');
      add('%%Creator: '+appname+' '+appversion);
      add('%%Title: '+extractfilename(PSfilename));
      if jdxrec1.title = '' then add('%%BoundingBox: 32 27 500 345')
        else add('%%BoundingBox: 32 27 500 355');
      add('gsave');
      add('0.6 setlinewidth');
      add('1 setlinecap');
      add('1 setlinejoin');
      add('/L '+cbo+'lineto'+cbc+' def');
      add('/M '+cbo+'moveto'+cbc+' def');
      add('/RL '+cbo+'rlineto'+cbc+' def');
      add('/RM '+cbo+'rmoveto'+cbc+' def');
      add('/XT '+cbo+'M -5 0 RL'+cbc+' def');
      add('/Yt '+cbo+'M 0 -3 RL'+cbc+' def');
      add('/YT '+cbo+'M 0 -6 RL'+cbc+' def');
      add('/RJ '+cbo+'stringwidth pop neg 0 RM'+cbc+' def');
      add('/CJ '+cbo+'stringwidth pop 2 div neg 0 RM'+cbc+' def');
      wstr := '';
      xstep := ((XindexHi-XindexLo) + 1) div (4*fullX);  // screen: 4*fullX
      if xstep < 1 then xstep := 1;
      if xstep > 4 then xstep := 8;
      if pdc = 1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexLo;
              endxindex   := XindexHi;
            end else
            begin
              startxindex := 1;
              endxindex   := jdxrec1.npoints;
            end;
          startXval   := XLo;
          endXval     := XHi;
        end;
      if pdc = -1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexHi;
              endxindex   := XindexLo;
            end else
            begin
              startxindex := jdxrec1.npoints;
              endxindex   := 1;
            end;
          startXval   := XHi;
          endXval     := XLo;
        end;
      // Grid
      if grid then
        begin
          add('%GRID');
          add('[1 2] 1 setdash');
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          wstr := '';
          for i := YscaleLo to YscaleHi do
            begin
              PSY := PSbaseY + ( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*PSfullY );
              if (PSY <= (PSbaseY+PSfullY)) and (PSY >= PSbaseY) then
                begin
                  //moveto(pixX,pixY);
                  //lineto(baseX,pixY);
                  str(PSbaseX:3:1,xstr);
                  str(PSY:3:1,ystr);
                  wstr := xstr+' '+ystr+' M ';
                  str((PSbaseX+PSfullX):3:1,xstr);
                  wstr := wstr + xstr + ' ' + ystr + ' L';
                  add(wstr);
                  wstr := '';
                  //textout(baseX-(10+textwidth(outstr)),(pixY-textheight(outstr) div 2),outstr);
                end;
            end;
          // vertical lines
          if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
          i := XscaleLo;
          while i <> XscaleHi do
            begin
              PSX := PSbaseX + abs((PSfullX * ((startXval - (i/Xscalefactor)) /
                      (abs(XHi - XLo) ) )));
              //moveto(pixX,pixY);
              str(PSX:3:1,xstr);
              str(PSbaseY:3:1,ystr);
              wstr := xstr+' '+ystr+' M ';
              str((PSbaseY+PSfullY):3:1,ystr);
              wstr := wstr + xstr + ' ' + ystr + ' L';
              add(wstr);
              wstr := '';
              i := i + xsdc;
            end;
          add('stroke [] 0 setdash');
          add('%ENDGRID');
        end;  // end of grid routine
      if datatype = dtXYdata then  // equally spaced X values
        begin
          add('%SPECTRUM (XYdata)');
          add('newpath');
          prevPSX := PSbaseX;
          prevPSY := PSbaseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*PSfullY );
          nval  := 0;
          Yaccu := 0;
          str(prevPSX:3:1,xstr);
          str(prevPSY:3:1,ystr);
          wstr := xstr+' '+ystr+' M ';
          //moveto(prevX,prevY);
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  PSX := PSbaseX + ( (abs(startxindex-xindex)/abs(XindexHi-XindexLo)) *PSfullX );
                  PSY := PSbaseY + ( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*PSfullY );
                  //lineto(pixX,pixY);
                  if (PSX <> prevPSX) or (PSY <> prevPSY) then
                    begin
                      str(PSX:3:1,xstr);
                      str(PSY:3:1,ystr);
                      wstr := wstr+xstr+' '+ystr+' L ';
                    end;
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevPSX := PSX;
                  prevPSY := PSY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('stroke');
        end;
      if datatype = dtXYpoints then  // non-equally-spaced X values
        begin
          add('%SPECTRUM (XYpoints)');
          add('newpath');
          prevPSX := PSbaseX;
          prevPSY := PSbaseY + ( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*PSfullY );
          nval  := 0;
          Yaccu := 0;
          //moveto(prevX,prevY);
          str(prevPSX:3:1,xstr);
          str(prevPSY:3:1,ystr);
          wstr := xstr+' '+ystr+' M ';
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  PSX := PSbaseX + abs((PSfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
                  PSY := PSbaseY + ( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*PSfullY );
                  //lineto(pixX,pixY);
                  if (PSX <> prevPSX) or (PSY <> prevPSY) then
                    begin
                      str(PSX:3:1,xstr);
                      str(PSY:3:1,ystr);
                      wstr := wstr + xstr+' '+ystr+' L ';
                    end;
                  if length(wstr)>60 then
                    begin
                      add(wstr);
                      wstr := '';
                    end;
                  prevPSX := PSX;
                  prevPSY := PSY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('stroke');
        end;
      if datatype = dtpeaktable then    // MS peak list etc.
        begin
          add('%SPECTRUM (peak table)');
          add('newpath');
          xindex := startxindex - pdc;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              PSX := PSbaseX + abs((PSfullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
              PSY := PSbaseY + ( ((xydata^[xindex].Y-minYadj)/((maxyadj-minYadj)))*PSfullY );
              if ((xydata^[xindex].X > XLo) and (xydata^[xindex].X < XHi)) or
                 ((xydata^[xindex].X < XLo) and (xydata^[xindex].X > XHi)) then
                begin
                  //moveto(pixX,baseY);
                  //lineto(pixX,pixY);
                  str(PSX:3:1,xstr);
                  str(PSbaseY:3:1,ystr);
                  wstr := xstr+' '+ystr+' M ';
                  str(PSY:3:1,ystr);
                  wstr := wstr + xstr+' '+ystr+' L';
                  add(wstr);
                  wstr := '';
                end;
            end;
          if wstr <> '' then
            begin
              add(wstr);
              wstr := '';
            end;
          add('stroke');
        end;
      // frame, X scale, Y scale, comment etc.
      add('%FRAME');
      str(PSbaseX:3:1,xstr);
      str(PSbaseY:3:1,ystr);
      add(xstr+' '+ystr+' M');
      str((PSbaseY+PSfullY):3:1,ystr);
      add(xstr+' '+ystr+' L');
      str((PSbaseX+PSfullX):3:1,xstr);
      add(xstr+' '+ystr+' L');
      str(PSbaseY:3:1,ystr);
      add(xstr+' '+ystr+' L');
      str(PSbaseX:3:1,xstr);
      add(xstr+' '+ystr+' L');
      add('stroke');
      add('%ENDFRAME');
      //moveto(baseX,baseY);
      //lineto(baseX+fullX,baseY);
      //lineto(baseX+fullX,baseY+fullY);
      //lineto(baseX,baseY+fullY);
      //lineto(baseX,baseY);
      // frame around spectrum
      // Y scale
      add('%SCALE');
      add('/Helvetica findfont 9 scalefont setfont');
      if spectype <> stNMR then
        begin
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          PSX := PSbaseX;
          for i := YscaleLo to YscaleHi do
            begin
              PSY := PSbaseY + ( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*PSfullY );
              if (PSY <= (PSbaseY+PSfullY)) and (PSY >= PSbaseY) then
                begin
                  //moveto(pixX,pixY);
                  //lineto(baseX,pixY);
                  Yscaleval := (i*Yscalecorr)/Yscalefactor;
                  str(Yscaleval:1:Yscaledigits,outstr);
                  str(PSbaseX:3:1,xstr);
                  str(PSY:3:1,ystr);
                  add(xstr+' '+ystr+' XT');
                  str((PSbaseX-10):3:1,xstr);
                  str((PSY-2):3:1,ystr);
                  add(xstr+' '+ystr+' M');
                  add('('+OutStr+') RJ');
                  add('('+OutStr+') show');
                  //textout(baseX-(10+textwidth(outstr)),(pixY-textheight(outstr) div 2),outstr);
                end;
            end;
          //textout(10, ((baseY+fullY) div 2) - (textheight(jdxrec1.yunits) div 2),jdxrec1.yunits);
          outstr := jdxrec1.yunits;
          str((PSbaseX/2):3:1,xstr);
          str((PSbaseY+PSfullY+14):3:1,ystr);
          add(xstr+' '+ystr+' M');
          add('('+OutStr+') show');
          wstr := '';
        end;
      // X scale
      PSY := PSbaseY;
      if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
      if abs(XscaleLo-XscaleHi) >= 100 then Xscalemod := 50 else
        if abs(XscaleLo-XscaleHi) >= 20 then Xscalemod := 10 else
          if abs(XscaleLo-XscaleHi) >= 5 then Xscalemod := 2 else Xscalemod := 1;
      Xscalemod := Xscalemod div round(fullX/600);  // should be derived from screen-fullX
      if Xscalemod < 1 then Xscalemod := 1;
      i := XscaleLo;
      str(PSbaseY:3:1,ystr);
      while i <> XscaleHi do
        begin
          PSX := PSbaseX + abs((PSfullX * ((startXval - (i/Xscalefactor)) /
                  (abs(XHi - XLo) ) )));
          if ((i mod Xscalemod) = 0) then
            begin
              //moveto(pixX,pixY+5);
              Xscaleval := (i/Xscalefactor);
              str(Xscaleval:1:Xscaledigits,outstr);
              str(PSX:3:1,xstr);
              str(PSbaseY:3:1,ystr);
              add(xstr+' '+ystr+' YT');
              str((PSbaseY-15):3:1,ystr);
              add(xstr+' '+ystr+' M');
              add('('+OutStr+') CJ');
              add('('+OutStr+') show');
            end else
            begin
              //moveto(pixX,pixY);
              str(PSX:3:1,xstr);
              str(PSbaseY:3:1,ystr);
              add(xstr+' '+ystr+' Yt');
            end;
          i := i + xsdc;
        end;
      //textout((baseX+fullX) - textwidth(jdxrec1.xunits), baseY+((paintbox1.height-baseY) div 2)-(textheight(jdxrec1.xunits) div 2),jdxrec1.xunits);
      outstr := jdxrec1.xunits;
      str((PSbaseX+PSfullX):3:1,xstr);
      str((PSbaseY/2.5):3:1,ystr);
      add(xstr+' '+ystr+' M');
      add('('+OutStr+') RJ');
      add('('+OutStr+') show');
      add('stroke');
      add('%ENDSCALE');
      // end of X/Y scale routine
      outstr := jdxrec1.title;
      if outstr <> '' then
        begin
          add('/Helvetica findfont 11 scalefont setfont');
          str((PSBaseX+(PSFullX/2)):3:1,xstr);
          str((PSBaseY+PSFullY+14):3:1,ystr);
          add(xstr+' '+ystr+' M');
          add('('+OutStr+') CJ');
          add('('+OutStr+') show');
        end;
      add('showpage');
      add('grestore');
      add(CtrlD);
      try
        savetofile(PSfilename);
      except
        messagedlg('Could not write Postscript file!',mtError,[mbOK],0);
      end;
    end;
end;


//========form functions and procedures============

procedure TForm1.Exit1Click(Sender: TObject);
begin
  close;
end;

procedure TForm1.Open1Click(Sender: TObject);
begin
  with OpenDialog1 do
    begin
      filename := '';
      Title := 'Open JCAMP-DX file';
      DefaultExt := '.jdx';
      // Win32 version:
      Filter := 'JCAMP-DX files (*.jdx;*.dx)|*.jdx;*.dx';
      // Linux version:
      //Filter := 'JCAMP-DX files (*.jdx;*.dx)';
    end;
  if OpenDialog1.Execute then
    begin
      specloaded := false;
      jdxfilename := OpenDialog1.FileName;
      readjdxfile(jdxfilename);
      paintbox1.Invalidate;  // required for Windows version
      if specloaded then
        begin
          caption := Appname + ' ' + AppVersion + '  -  '+jdxfilename;
          save1.enabled         := true;
          plot1.enabled         := true;
          parameters1.enabled   := true;
          grid1.enabled         := true;
          grid1.Checked         := false;
          grid                  := false;
          clipspectrum1.enabled := true;
          button1.enabled       := true;
          button2.enabled       := true;
          button3.enabled       := true;
          button4.enabled       := true;
          button5.enabled       := true;
        end else
        begin
          caption := Appname + ' ' + AppVersion;
          save1.enabled         := false;
          plot1.enabled         := false;
          parameters1.enabled   := false;
          clipspectrum1.enabled := false;
          button1.enabled       := false;
          button2.enabled       := false;
          button3.enabled       := false;
          button4.enabled       := false;
          button5.enabled       := false;
          Statusbar1.Panels[0].text := '';
          Statusbar1.Panels[1].text := 'X1:';
          Statusbar1.Panels[2].text := 'Y1:';
          Statusbar1.Panels[3].text := 'X2:';
          Statusbar1.Panels[4].text := 'Y2:';
          Statusbar1.Panels[5].text := 'delta X:';
        end;
    end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  pstr : string;
begin
  application.title := 'JDXview';
  Caption := Appname + ' ' + AppVersion;
  scrollbox1.Height := form1.ClientHeight - 22;
  scrollbox1.width := form1.ClientWidth - 43;
  paintbox1.width  := scrollbox1.width + 1;
  paintbox1.Height := scrollbox1.Height + 1;
  jdxstringlist1 := Tstringlist.Create;
  jdxstringlist2 := Tstringlist.Create;
  OffsetLeft := 50;
  OffsetBottom := 70;
  LcursX := OffsetLeft;
  RcursX := OffsetLeft;
  pdc := 1;
  pdc := 1; // standard plot direction
  Statusbar1.Panels[1].text := 'X1:';
  Statusbar1.Panels[2].text := 'Y1:';
  Statusbar1.Panels[3].text := 'X2:';
  Statusbar1.Panels[4].text := 'Y2:';
  Statusbar1.Panels[5].text := 'delta X:';
  Statusbar1.Panels[6].text := '';
  save1.enabled         := false;
  plot1.enabled         := false;
  parameters1.enabled   := false;
  grid1.Checked         := false;
  grid1.enabled         := false;
  clipspectrum1.enabled := false;
  button1.enabled       := false;
  button2.enabled       := false;
  button3.enabled       := false;
  button4.enabled       := false;
  button5.enabled       := false;
  datamode              := dmASDF;
  grid                  := false;
  try
    GetMem(xydata,SizeOf(xyarray));
  except
    on e:EOutOfMemory do
      begin
        MessageDlg('Not enough memory',mtError,[mbOK],0);
        Close;
      end;
  end;
  if paramcount > 0 then
    begin
      pstr := paramstr(paramcount);
      if pos('-',pstr) <> 1 then
        begin
          if not fileexists(pstr) then
              MessageDlg('File '+pstr+' not found!',mtError,[mbOK],0) else
            begin
              Statusbar1.Panels[6].text := pstr;
              readjdxfile(pstr);
              paintbox1.Invalidate;
              if specloaded then
                begin
                  jdxfilename           := pstr;
                  caption := Appname + ' ' + AppVersion + '  -  '+jdxfilename;
                  save1.enabled         := true;
                  plot1.enabled         := true;
                  parameters1.enabled   := true;
                  grid1.enabled         := true;
                  grid1.Checked         := false;
                  grid                  := false;
                  clipspectrum1.enabled := true;
                  button1.enabled       := true;
                  button2.enabled       := true;
                  button3.enabled       := true;
                  button4.enabled       := true;
                  button5.enabled       := true;
                end;
            end;
        end;
    end;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  {Constraints}
  if Form1.Width  < 300 then Form1.Width  := 300;
  if Form1.Height < 300 then Form1.Height := 300;
  bevel1.width := form1.ClientWidth;
  scrollbox1.Height := form1.ClientHeight - 22;
  scrollbox1.width := form1.ClientWidth - 43;
  if paintbox1.width < (scrollbox1.width + 1) then
     paintbox1.width  := scrollbox1.width + 1;
  if paintbox1.Height < (scrollbox1.Height + 1) then
     paintbox1.Height := scrollbox1.Height + 1;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  try
    if xydata <> nil then FreeMem(xydata,SizeOf(xyarray));
  except
    on e:EInvalidPointer do begin end;
  end;
  jdxstringlist1.free;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  if not specloaded then exit;
  if abs(LcursXval-RcursXval) >= 0.005*abs(iniXLo-iniXHi) then
    begin
      if pdc = 1 then
        begin
          if LcursX < RcursX then
            begin
              XLo := LcursXval;
              XHi := RcursXval;
              XindexLo := LCursxindex;
              XindexHi := RcursXindex;
            end else
            begin
              XLo := RcursXval;
              XHi := LcursXval;
              XindexLo := RCursxindex;
              XindexHi := LcursXindex;
            end;
        end else
        begin
          if LcursX < RcursX then
            begin
              XLo := RcursXval;
              XHi := LcursXval;
              XindexLo := RCursxindex;
              XindexHi := LcursXindex;
            end else
            begin
              XLo := LcursXval;
              XHi := RcursXval;
              XindexLo := LCursxindex;
              XindexHi := RcursXindex;
            end;
        end;
      adjustminmaxX;
      paintbox1.invalidate;
    end;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  if not specloaded then exit;
  XLo := iniXlo;
  XHi := iniXHi;
  XindexLo := 1;
  XindexHi := jdxrec1.npoints;;
  adjustminmaxX;
  paintbox1.width := scrollbox1.width + 5;
  paintbox1.invalidate;
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  if not specloaded then exit;
  if paintbox1.height < 8000 then
    begin
      paintbox1.height := paintbox1.height * 2;
      scrollbox1.vertscrollbar.position := scrollbox1.vertscrollbar.range;
      paintbox1.refresh;
    end;
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
  if not specloaded then exit;
  if paintbox1.height/2 < scrollbox1.height + 5 then
     paintbox1.Height := scrollbox1.Height +5 else
    paintbox1.height := paintbox1.height div 2;
  paintbox1.refresh;
end;

procedure TForm1.DrawCursor(CX : integer; cno : integer);
var
  pbcol  : tcolor;
  pbmode : tpenmode;
begin
  if not specloaded then exit;
  pbcol  := paintbox1.canvas.pen.color;
  pbmode := paintbox1.canvas.pen.mode;
  case cno of
    1 : paintbox1.canvas.pen.color := clRed;
    2 :  paintbox1.canvas.pen.color := clLime;
  end;
  paintbox1.canvas.pen.mode  := pmNotXor;
  PaintBox1.Canvas.MoveTo(CX, BaseY-1);
  PaintBox1.Canvas.Lineto(CX, BaseY+FullY);
  paintbox1.canvas.pen.color := pbcol;
  paintbox1.canvas.pen.mode  := pbmode;
end;


procedure TForm1.PaintBox1Paint(Sender: TObject);
var
  PrevX, PrevY, PixX, PixY, auxint, i  : integer;
  aux1 : real;
  auxreal : real;
  outstr  : string;
  ScanCM  : real;
  s : integer;
  xindex, startxindex, endxindex : integer;
  Yaccu : extended;
  xstep, nval : integer;
  Yscaleval : double;
  Yscalecorr : integer;
  Xscaleval : double;
  xsdc : integer;
  startXval : double;
  endXval   : double;
  frame : trect;
begin
  if nographics then exit;
  if not specloaded then exit;
  with PaintBox1.canvas do
    begin
      pen.color := clBlack;
      pen.mode  := pmCopy;
      pen.style := psSolid;
      font.name := 'Arial Narrow';
      font.height := -12;
      MaxX := PaintBox1.Width-30;
      MaxY := PaintBox1.Height-20;
      BaseX := OffsetLeft;
      BaseY := PaintBox1.Height - OffsetBottom;
      FullX := MaxX-(OffsetLeft + 2);
      FullY := -(maxY - (OffsetBottom + 10));
      xstep := ((XindexHi-XindexLo) + 1) div (4*fullX);
      if xstep < 1 then xstep := 1;
      if xstep > 4 then xstep := 8;
      if pdc = 1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexLo;
              endxindex   := XindexHi;
            end else
            begin
              startxindex := 1;
              endxindex   := jdxrec1.npoints;
            end;
          startXval   := XLo;
          endXval     := XHi;
        end;
      if pdc = -1 then
        begin
          if spectype <> stMS then
            begin
              startxindex := XindexHi;
              endxindex   := XindexLo;
            end else
            begin
              startxindex := jdxrec1.npoints;
              endxindex   := 1;
            end;
          startXval   := XHi;
          endXval     := XLo;
        end;
      // Grid
      if grid then
        begin
          pen.color := clSilver;
          pen.style := psDot;
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          for i := YscaleLo to YscaleHi do
            begin
              pixY := baseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*fullY );
              if (pixY >= (baseY+fullY)) and (pixY <= baseY) then
                begin
                  moveto(baseX,pixY);
                  lineto(baseX+fullX,pixY);
                end;
            end;
          // vertical lines
          pixY := baseY + 5;
          if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
          if abs(XscaleLo-XscaleHi) >= 100 then Xscalemod := 50 else
            if abs(XscaleLo-XscaleHi) >= 20 then Xscalemod := 10 else
              if abs(XscaleLo-XscaleHi) >= 5 then Xscalemod := 2 else Xscalemod := 1;
          Xscalemod := Xscalemod div round(fullX/600);
          if Xscalemod < 1 then Xscalemod := 1;
          i := XscaleLo;
          while i <> XscaleHi do
            begin
              pixX := baseX + abs(round(fullX * ((startXval - (i/Xscalefactor)) /
                      (abs(XHi - XLo) ) )));
              moveto(pixX,baseY);
              lineto(pixX,baseY+fullY);
              i := i + xsdc;
            end;
          // end grid routine
          pen.color := clBlack;
          pen.style := psSolid;
        end;
      // Frame
      //frame.left := baseX;
      //frame.Right := baseX+fullX;
      //frame.bottom := baseY;
      //frame.Top := baseY+fullY;
      // frame routine for Win32
      //framerect(frame);
      brush.Style := bsClear;
      // frame routine for Kylix
      moveto(baseX,baseY);
      lineto(baseX+fullX,baseY);
      lineto(baseX+fullX,baseY+fullY);
      lineto(baseX,baseY+fullY);
      lineto(baseX,baseY);
      // Y scale
      if spectype <> stNMR then
        begin
          if (specsubtype = sstIRtransmittance) and (maxYadj <= 1) then
            Yscalecorr := 100 else Yscalecorr := 1;
          pixX := baseX - 5;
          for i := YscaleLo to YscaleHi do
            begin
              pixY := baseY+round( (((i/Yscalefactor)-minYadj)/((maxYadj-minYadj)))*fullY );
              if (pixY >= (baseY+fullY)) and (pixY <= baseY) then
                begin
                  moveto(pixX,pixY);
                  lineto(baseX,pixY);
                  Yscaleval := (i*Yscalecorr)/Yscalefactor;
                  str(Yscaleval:1:Yscaledigits,outstr);
                  textout(baseX-(10+textwidth(outstr)),(pixY-textheight(outstr) div 2),outstr);
                end;
            end;
          textout(10, ((baseY+fullY) div 2) - (textheight(jdxrec1.yunits) div 2),jdxrec1.yunits);
        end;
      // X scale
      pixY := baseY + 5;
      if XscaleLo > XscaleHi then xsdc := -1 else xsdc := 1;
      if abs(XscaleLo-XscaleHi) >= 100 then Xscalemod := 50 else
        if abs(XscaleLo-XscaleHi) >= 20 then Xscalemod := 10 else
          if abs(XscaleLo-XscaleHi) >= 5 then Xscalemod := 2 else Xscalemod := 1;
      Xscalemod := Xscalemod div round(fullX/600);
      if Xscalemod < 1 then Xscalemod := 1;
      i := XscaleLo;
      while i <> XscaleHi do
        begin
          pixX := baseX + abs(round(fullX * ((startXval - (i/Xscalefactor)) /
                  (abs(XHi - XLo) ) )));
          if ((i mod Xscalemod) = 0) then moveto(pixX,pixY+5) else moveto(pixX,pixY);
          lineto(pixX,baseY);
          Xscaleval := (i/Xscalefactor);
          str(Xscaleval:1:Xscaledigits,outstr);
          if ((i mod Xscalemod) = 0) then textout(pixX-(textwidth(outstr) div 2),
                                          (pixY+textheight(outstr) - 8),outstr);
          i := i + xsdc;
        end;
      textout((baseX+fullX) - textwidth(jdxrec1.xunits),
               baseY+((paintbox1.height-baseY) div 2)-(textheight(jdxrec1.xunits) div 2),jdxrec1.xunits);
      // end of X/Y scale routine
      // title
      textout(((baseX+fullX) div 2) - (textwidth(jdxrec1.title) div 2),
              ((baseY+fullY) div 2) - (textheight(jdxrec1.title) div 2),jdxrec1.title);
      // end title
      if datatype = dtXYdata then  // equally spaced X values
        begin
          prevX := baseX;
          prevY := baseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*fullY );
          nval  := 0;
          Yaccu := 0;
          moveto(prevX,prevY);
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := baseX + round( (abs(startxindex-xindex)/abs(XindexHi-XindexLo)) *fullX );
                  pixY := baseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*fullY );
                  lineto(pixX,pixY);
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
        end;
      if datatype = dtXYpoints then  // non-equally-spaced X values
        begin
          prevX := baseX;
          prevY := baseY+round( ((xydata^[startxindex].Y-minYadj)/((maxYadj-minYadj)))*fullY );
          nval  := 0;
          Yaccu := 0;
          moveto(prevX,prevY);
          xindex := startxindex;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              Yaccu := Yaccu + xydata^[xindex].Y;
              inc(nval);
              if (nval = xstep) or (xindex = endxindex) then
                begin
                  pixX := baseX + abs(round(fullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
                  pixY := baseY+round( (((Yaccu/nval)-minYadj)/((maxyadj-minYadj)))*fullY );
                  lineto(pixX,pixY);
                  prevX := pixX;
                  prevY := pixY;
                  Yaccu := 0;
                  nval := 0;
                end;
            end;
        end;
      if datatype = dtpeaktable then    // MS peak list etc.
        begin
          xindex := startxindex - pdc;
          while xindex <> endxindex do
            begin
              xindex := xindex + pdc;
              pixX := baseX + abs(round(fullX * ((startXval - xydata^[xindex].X) / (abs(XHi - XLo) ) )));
              pixY := baseY+round( ((xydata^[xindex].Y-minYadj)/((maxyadj-minYadj)))*fullY );
              if ((xydata^[xindex].X > XLo) and (xydata^[xindex].X < XHi)) or
                 ((xydata^[xindex].X < XLo) and (xydata^[xindex].X > XHi)) then
                begin
                  moveto(pixX,baseY);
                  lineto(pixX,pixY);
                end;
            end;
        end;
    // draw the cursor
      if pdc = 1 then
        begin
          LcursX := baseX + round(fullX*( (XLo-LcursXval)/(XLo-XHi)   ) );
          RcursX := baseX + round(fullX*( (XLo-RcursXval)/(XLo-XHi)   ) );
        end else
        begin
          LcursX := baseX + fullX - round(fullX*( (XLo-LcursXval)/(XLo-XHi)   ) );
          RcursX := baseX + fullX - round(fullX*( (XLo-RcursXval)/(XLo-XHi)   ) );
        end;
      if (LcursX > BaseX) and (LcursX < BaseX+FullX) then DrawCursor(LcursX, 1);
      str(LcursXval:1:Xscaledigits,outstr);
      statusbar1.panels[1].text := 'X1: '+outstr;
      str(LcursYval:1:Yscaledigits+2,outstr);
      statusbar1.panels[2].text := 'Y1: '+outstr;
      if (RcursX > BaseX) and (RcursX < BaseX+FullX) then DrawCursor(RcursX, 2);
      str(RcursXval:1:Xscaledigits,outstr);
      statusbar1.panels[3].text := 'X2: '+outstr;
      str(RcursYval:1:Yscaledigits+2,outstr);
      statusbar1.panels[4].text := 'Y2: '+outstr;
      str(abs(LcursXval-RcursXval):1:Xscaledigits,outstr);
      Statusbar1.Panels[5].text := 'delta X: '+outstr;
      if showdeltaHz then
        begin
          str(deltaHz:1:2,outstr);
          Statusbar1.Panels[6].text := 'delta X (Hz): '+outstr;
        end else Statusbar1.Panels[6].text := '';
    end;
end;

procedure TForm1.Parameters1Click(Sender: TObject);
begin
  if jdxstringlist1.Capacity = 0 then exit;
  form2 := tform2.Create(self);
  if form2.ShowModal = mrOK then
    begin
    end;
end;

procedure TForm1.Button5Click(Sender: TObject);
// revert plot direction
var
  pdctmp : integer;
begin
  if not specloaded then exit;
  if pdc = 1  then pdctmp := -1;
  if pdc = -1 then pdctmp := 1;
  pdc := pdctmp;
  paintbox1.Invalidate;
end;

procedure TForm1.DumpXYdata1Click(Sender: TObject);
begin
  dumpXYvalues('jdxdump.txt');
end;

procedure TForm1.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  outstr : string;
begin
  if not specloaded then exit;
  if button = mbLeft then
    begin
      if (X+1 < ScrollBox1.HorzScrollBar.Position) or
         (X+1 >= ScrollBox1.Width+ScrollBox1.HorzScrollBar.Position) or
         (Y > BaseY) or (Y < BaseY+FullY) then
        begin
          InScrollBox := false;
          exit;
        end else InScrollBox := true;
      Lcdrawing := true;
      if pdc = 1 then
        begin
          LcursX := baseX + round(fullX*( (XLo-LcursXval)/(XLo-XHi)   ) );
        end else
        begin
          LcursX := baseX + fullX - (round(fullX*( (XLo-LcursXval)/(XLo-XHi)   ) ));
        end;
      if (LcursX > BaseX) and (LcursX < BaseX+FullX) then DrawCursor(LcursX, 1);
      if (X >= baseX) and (X <= (baseX+fullX)) then LcursX := X;
      if pdc = 1 then
        begin
          LcursXval := XLo + ((LcursX-baseX)/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              LcursXindex := XindexLo + round( ((LcursX-baseX)/(fullX))*(XindexHi-XindexLo));
            end;
        end else
        begin
          LcursXval := XLo - ((LcursX-(baseX+fullX))/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              LcursXindex := XindexLo - round( ((LcursX-(baseX+fullX))/(fullX))*(XindexHi-XindexLo));
            end;
        end;
      if (LcursX > BaseX) and (LcursX < BaseX+FullX) then DrawCursor(LcursX, 1);
      if (LcursX >= BaseX) and (LcursX <= BaseX+FullX) then
        begin
          str(LcursXval:1:Xscaledigits,outstr);
          statusbar1.panels[1].text := 'X1: '+outstr;
          str(abs(LcursXval-RcursXval):1:Xscaledigits,outstr);
          Statusbar1.Panels[5].text := 'delta X: '+outstr;
        end;
    end;
  if button = mbRight then
    begin
      if (X+1 < ScrollBox1.HorzScrollBar.Position) or
         (X+1 >= ScrollBox1.Width+ScrollBox1.HorzScrollBar.Position) or
         (Y > BaseY) or (Y < BaseY+FullY) then
        begin
          InScrollBox := false;
          exit;
        end else InScrollBox := true;
      Rcdrawing := true;
      if pdc = 1 then
        begin
          RcursX := baseX + round(fullX*( (XLo-RcursXval)/(XLo-XHi)   ) );
        end else
        begin
          RcursX := baseX + fullX - (round(fullX*( (XLo-RcursXval)/(XLo-XHi)   ) ));
        end;
      if (RcursX > BaseX) and (RcursX < BaseX+FullX) then DrawCursor(RcursX, 2);
      if (X >= baseX) and (X <= (baseX+fullX)) then RcursX := X;
      if pdc = 1 then
        begin
          RcursXval := XLo + ((RcursX-baseX)/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              RcursXindex := XindexLo + round( ((RcursX-baseX)/(fullX))*(XindexHi-XindexLo));
            end;
        end else
        begin
          RcursXval := XLo -  ((RcursX-(baseX+fullX))/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              RcursXindex := XindexLo - round( ((RcursX-(baseX+fullX))/(fullX))*(XindexHi-XindexLo));
            end;
        end;
      if (RcursX > BaseX) and (RcursX < BaseX+FullX) then DrawCursor(RcursX, 2);
      if (RcursX >= BaseX) and (RcursX <= BaseX+FullX) then
        begin
          str(RcursXval:1:Xscaledigits,outstr);
          statusbar1.panels[3].text := 'X2: '+outstr;
          str(abs(LcursXval-RcursXval):1:Xscaledigits,outstr);
          Statusbar1.Panels[5].text := 'delta X: '+outstr;
        end;
    end;
end;

procedure TForm1.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  outstr : string;
begin
  if not specloaded then exit;
  if Lcdrawing then
    begin
      if (X+1 < ScrollBox1.HorzScrollBar.Position) or
         (X+1 >= ScrollBox1.Width+ScrollBox1.HorzScrollBar.Position) or
         (Y > BaseY) or (Y < BaseY+FullY) then
        begin
          InScrollBox := false;
          exit;
        end else InScrollBox := true;
      if pdc = 1 then
        begin
          LcursX := baseX + round(fullX*( (XLo-LcursXval)/(XLo-XHi) ));
        end else
        begin
          LcursX := baseX + fullX - (round(fullX*( (XLo-LcursXval)/(XLo-XHi) )));
        end;
      if (LcursX > BaseX) and (LcursX < BaseX+FullX) then DrawCursor(LcursX, 1);
      if (X >= baseX) and (X <= (baseX+fullX)) then LcursX := X;
      if X < baseX then LcursX := baseX;
      if X > (baseX+fullX) then LcursX := baseX+fullX;
      if pdc = 1 then
        begin
          LcursXval := XLo + ((LcursX-baseX)/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              LcursXindex := XindexLo + round( ((LcursX-baseX)/(fullX))*(XindexHi-XindexLo));
            end;
        end else
        begin
          LcursXval := XLo - ((LcursX-(baseX+fullX))/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              LcursXindex := XindexLo - round( ((LcursX-(baseX+fullX))/(fullX))*(XindexHi-XindexLo));
            end;
        end;
      if (LcursX > BaseX) and (LcursX < BaseX+FullX) then DrawCursor(LcursX, 1);
      if (LcursX >= BaseX) and (LcursX <= BaseX+FullX) then
        begin
          str(LcursXval:1:Xscaledigits,outstr);
          statusbar1.panels[1].text := 'X1: '+outstr;
          str(abs(LcursXval-RcursXval):1:Xscaledigits,outstr);
          Statusbar1.Panels[5].text := 'delta X: '+outstr;
        end;
      statusbar1.panels[2].text := 'Y1: ';
      if showdeltaHz then Statusbar1.Panels[6].text := 'delta X (Hz): ';
    end;
  if Rcdrawing then
    begin
      if (X+1 < ScrollBox1.HorzScrollBar.Position) or
         (X+1 >= ScrollBox1.Width+ScrollBox1.HorzScrollBar.Position) or
         (Y > BaseY) or (Y < BaseY+FullY) then
        begin
          InScrollBox := false;
          exit;
        end else InScrollBox := true;
      if pdc = 1 then
        begin
          RcursX := baseX + round(fullX*( (XLo-RcursXval)/(XLo-XHi) ));
        end else
        begin
          RcursX := baseX + fullX - (round(fullX*( (XLo-RcursXval)/(XLo-XHi) )));
        end;
      if (RcursX > BaseX) and (RcursX < BaseX+FullX) then DrawCursor(RcursX, 2);
      if (X >= baseX) and (X <= (baseX+fullX)) then RcursX := X;
      if X < baseX then RcursX := baseX;
      if X > (baseX+fullX) then RcursX := baseX+fullX;
      if pdc = 1 then
        begin
          RcursXval := XLo + ((RcursX-baseX)/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              RcursXindex := XindexLo + round( ((RcursX-baseX)/(fullX))*(XindexHi-XindexLo));
            end;
        end else
        begin
          RcursXval := XLo -  ((RcursX-(baseX+fullX))/(fullX))*(XHi-XLo);
          if not ((spectype = stMS) and (specsubtype = sstMSpeaklist)) then
            begin
              RcursXindex := XindexLo - round( ((RcursX-(baseX+fullX))/(fullX))*(XindexHi-XindexLo));
            end;
        end;
      if (RcursX > BaseX) and (RcursX < BaseX+FullX) then DrawCursor(RcursX, 2);
      if (RcursX >= BaseX) and (RcursX <= BaseX+FullX) then
        begin
          str(RcursXval:1:Xscaledigits,outstr);
          statusbar1.panels[3].text := 'X2: '+outstr;
          str(abs(LcursXval-RcursXval):1:Xscaledigits,outstr);
          Statusbar1.Panels[5].text := 'delta X: '+outstr;
        end;
      statusbar1.panels[4].text := 'Y2: ';
      if showdeltaHz then Statusbar1.Panels[6].text := 'delta X (Hz): ';
    end;
end;

procedure TForm1.PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  cindex : integer;
  outstr : string;
begin
  if not specloaded then exit;
  if button = mbLeft then
    begin
      Lcdrawing := false;
      cindex := findXindex(LcursXval);
      if (cindex < 1) or (cindex > jdxrec1.npoints) then
        begin
          //debugoutput('cursor index out of range: '+inttostr(cindex));
          LcursYval := 0;
        end else
        begin
          LcursYval := xydata^[cindex].Y;
        end;
      str(LcursYval:1:Yscaledigits+2,outstr);
      statusbar1.panels[2].text := 'Y1: '+outstr;
    end;
  if button = mbRight then
    begin
      Rcdrawing := false;
      cindex := findXindex(RcursXval);
      if (cindex < 1) or (cindex > jdxrec1.npoints) then
        begin
          //debugoutput('cursor index out of range: '+inttostr(cindex));
          RcursYval := 0;
        end else
        begin
          RcursYval := xydata^[cindex].Y;
        end;
      str(RcursYval:1:Yscaledigits+2,outstr);
      statusbar1.panels[4].text := 'Y2: '+outstr;
    end;
  if showdeltaHz then
    begin
      deltaHz := abs(LcursXval-RcursXval)*SFO1;
      str(deltaHz:1:2,outstr);
      Statusbar1.Panels[6].text := 'delta X (Hz): '+outstr;
    end;
end;

procedure TForm1.Nextpage1Click(Sender: TObject);
begin
  if foundntuples then inc(page);
  readjdx;
  paintbox1.invalidate;
end;

procedure TForm1.Save1Click(Sender: TObject);
begin
  if form1.nextpage1.visible then   // new in v0.2
    begin
      if (messagedlg('Only the current page will be saved!',
        mtWarning,[mbOK,mbCancel],0) <> mrOK) then exit;
    end;
  with savedialog1 do
    begin
      defaultext := 'jdx';
      //for Win32:
      filter     := 'JCAMP-DX (*.jdx)|*.jdx|Tab-separated XY values (*.txt)|*.txt';
      //for Kylix:
      //filter     := 'JCAMP-DX (*.jdx)|Tab-separated XY values (*.txt)';
      title      := 'Save current spectrum as...';
      if jdxfilename <> '' then filename := jdxfilename;
    end;
  if SaveDialog1.Execute then
    begin
      if fileexists(savedialog1.filename) then
        begin
          if messagedlg('File exists! Overwrite?',mtConfirmation,[mbYes,mbNo],0)
           <> mrYes then exit;
        end;
      if savedialog1.filterindex = 1 then writejdxfile(savedialog1.filename);
      if savedialog1.filterindex = 2 then dumpXYvalues(savedialog1.filename);
    end;
end;

procedure TForm1.Clipspectrum1Click(Sender: TObject);
var
  loclipindex, hiclipindex : longint;
  auxint : longint;
begin
  str(LcursXval:1:Xscaledigits,loclipstr);
  str(RcursXval:1:Xscaledigits,hiclipstr);
  form3.loclipedit.text := loclipstr;
  form3.hiclipedit.text := hiclipstr;
  if form3.showmodal = mrOK then
    begin
      if (datatype <> dtXYpoints) and (datatype <> dtpeaktable) then
        begin
          if loclipval < minXval then loclipval := minXval;
          if loclipval > maxXval then loclipval := maxXval;
          if hiclipval < minXval then hiclipval := minXval;
          if hiclipval > maxXval then hiclipval := maxXval;
          loclipindex := 1 + round( ((xydata^[1].X-loclipval)/(xydata^[1].X-xydata^[jdxrec1.npoints].X))*(jdxrec1.npoints -1));
          hiclipindex := 1 + round( ((xydata^[1].X-hiclipval)/(xydata^[1].X-xydata^[jdxrec1.npoints].X))*(jdxrec1.npoints -1));
          if loclipindex > hiclipindex then
            begin
              auxint := loclipindex;
              loclipindex := hiclipindex;
              hiclipindex := auxint;
            end;
          //debugoutput('Lower clipping index: '+inttostr(loclipindex)+'  upper clipping index: '+inttostr(hiclipindex));
          if (not ((loclipindex = 1) and (hiclipindex = jdxrec1.npoints))) and
             (hiclipindex-loclipindex > 0) then
             begin
               screen.cursor := crHourglass;
               clipbyindex(loclipindex,hiclipindex);
               adjustminmax;
               screen.cursor := crDefault;
               paintbox1.invalidate;
             end;
        end else
        begin
          screen.cursor := crHourglass;
          clipbyval(loclipval,hiclipval);
          adjustminmax;
          screen.cursor := crDefault;
          paintbox1.invalidate;
        end;
    end;
end;

procedure TForm1.About1Click(Sender: TObject);
begin
  messagedlg(appname+' '+appversion+#13+'N. Haider  2003-2012'+#13+'norbert.haider@univie.ac.at',mtInformation,[mbOK],0);
end;

procedure TForm1.HPGLfile1Click(Sender: TObject);
begin
  // write a HPGL plot file
  with savedialog1 do
    begin
      defaultext := 'plt';
      filter     := 'HPGL plot file (*.plt)|*.plt';
      title      := 'Save plot file as...';
      if jdxfilename <> '' then filename := changefileext(jdxfilename,'.plt');
    end;
  if SaveDialog1.Execute then
    begin
      if fileexists(savedialog1.filename) then
        begin
          if messagedlg('File exists! Overwrite?',mtConfirmation,[mbYes,mbNo],0)
           <> mrYes then exit;
        end;
      writepltfile(savedialog1.filename);
    end;
end;

procedure TForm1.SVGfile1Click(Sender: TObject);
begin
  // write a Scalable Vector Graphics (SVG) file
  with savedialog1 do
    begin
      defaultext := 'svg';
      filter     := 'Scalable Vector Graphics file (*.svg)|*.svg';
      title      := 'Save SVG file as...';
      if jdxfilename <> '' then filename := changefileext(jdxfilename,'.svg');
    end;
  if SaveDialog1.Execute then
    begin
      if fileexists(savedialog1.filename) then
        begin
          if messagedlg('File exists! Overwrite?',mtConfirmation,[mbYes,mbNo],0)
           <> mrYes then exit;
        end;
      writesvgfile(savedialog1.filename);
    end;
end;

procedure TForm1.EPSfile1Click(Sender: TObject);
begin
  // write an Encapsulated Postscript fileend;
  with savedialog1 do
    begin
      defaultext := 'eps';
      filter     := 'Encapsulated Postscript file (*.eps)|*.eps';
      title      := 'Save Postscript file as...';
      if jdxfilename <> '' then filename := changefileext(jdxfilename,'.eps');
    end;
  if SaveDialog1.Execute then
    begin
      if fileexists(savedialog1.filename) then
        begin
          if messagedlg('File exists! Overwrite?',mtConfirmation,[mbYes,mbNo],0)
           <> mrYes then exit;
        end;
      writepsfile(savedialog1.filename);
    end;
end;

procedure TForm1.Grid1Click(Sender: TObject);
begin
  Grid1.Checked := not Grid1.Checked;
  if Grid1.Checked then grid := true else grid := false;
  PaintBox1.invalidate;
end;

end.
