unit jdxUnit2;
(*
This software is published under the terms of the GNU General Public
License (GPL, see below). For a detailed description of this license,
see http://www.gnu.org/copyleft/gpl.html

  JDXview - a utility program for displaying spectra in JCAMP-DX format.
  Author: Dr. Norbert Haider, University of Vienna, Department of Drug
  Synthesis, Althanstrasse 14, A-1090 Vienna, Austria.

  Inspired by Dr. Robert Lancashire's JCAMP-DX data viewer and its
  reincarnation in the MDL Chime plug-in, JDXview was developed as an
  external "helper application" which can be used in combination with
  web browsers, if they do not have a suitable plug-in installed.
  JDXview displays various kinds of spectra (NMR, IR, MS, etc.) in
  JCAMP-DX format. Supported formats are: uncompressed (AFFN) and
  compressed (ASDF, e.g. DIFDUP) as described in the 4.24 specification.
  Some features of more recent JCAMP-DX versions are also implemented,
  but not all of them.

  Furthermore, JDXview offers also rudimentary processing options: the
  range of a spectrum can be clipped, and the result saved in JCAMP-DX
  format again. Zooming and measuring of distances can be performed by
  two cursors which are placed by pressing the left or right mouse button,
  respectively, followed by a click on the "< >" button.

  Graphics output is supported in the following vector graphics formats:
  HP-GL (Hewlett-Packard Graphics Language)
  SVG (Scalable Vector Graphics)
  EPS (Encapsulated PostScript)

===============================================================================
DISCLAIMER
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
===============================================================================
*)


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids;

type
  TForm2 = class(TForm)
    StringGrid1: TStringGrid;
    OKbtn: TButton;
    procedure OKbtnClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form2: TForm2;

implementation

uses
  jdxUnit1;

{$R *.DFM}

procedure TForm2.OKbtnClick(Sender: TObject);
begin
  close;
end;

procedure TForm2.FormShow(Sender: TObject);
var
  i : integer;
  tmpstr : string;
begin
  if jdxstringlist1.Capacity = 0 then exit;
  stringgrid1.cells[0,0] := 'Parameter';
  stringgrid1.cells[1,0] := 'Value';
  i := -1;
  tmpstr := '';
  repeat
    begin
      inc(i);
      tmpstr := jdxstringlist1.strings[i];
      if pos('##',tmpstr)=1 then delete(tmpstr,1,2);
      stringgrid1.RowCount := i + 2;
      stringgrid1.cells[0,(i+1)] := copy(tmpstr,1,pos('=',tmpstr)-1);
      tmpstr := copy(tmpstr,pos('=',tmpstr)+1,length(tmpstr)-pos('=',tmpstr));
      stringgrid1.cells[1,(i+1)] := trimleft(tmpstr);
    end;
  until (i = jdxstringlist1.count-1) or (pos('##XYDATA',jdxstringlist1.strings[i])=1)
                                     or (pos('##XYPOINT',jdxstringlist1.strings[i])=1)
                                     or (pos('##PEAK',jdxstringlist1.strings[i])=1)
                                     or (pos('##NTUPLES',jdxstringlist1.strings[i])=1);
end;

procedure TForm2.FormCreate(Sender: TObject);
begin
  stringgrid1.defaultcolwidth := stringgrid1.width div 2;
end;

end.
